/* Ankur Parikh
   CondensedDMat.cpp

   see .hpp for description
   */


#include "CondensedDMat.hpp"

using namespace Eigen;
using namespace std;


CondensedDMat::CondensedDMat()
{
    index_map = NULL;
    mode = -1;
    cond_mat = NULL;
}

CondensedDMat::~CondensedDMat()
{
    delete(index_map);
    delete(cond_mat);
}

CondensedDMat::CondensedDMat(MatrixXd& m, map<int, int>& index_map, int mode)
{
    this->cond_mat = new MatrixXd(m);
    this->index_map = new map<int, int>(index_map);
    this->mode = mode;
}


double CondensedDMat::At(int i, int j)
{
    if (mode == 0)
    {
        int r = (*index_map)[i];
        return cond_mat->coeff(r, j);
    }
    else if (mode == 1)
    {
        int c = (*index_map)[j];
        return cond_mat->coeff(i, c);
    }
    else
        assert(0);
}


VectorXd CondensedDMat::Row(int i)
{
    assert(mode == 0);

    int r = (*index_map)[i];
    return cond_mat->row(r);
}

VectorXd CondensedDMat::Col(int j)
{
    assert(mode == 1);
    int c = (*index_map)[j];
    return cond_mat->col(c);
}

bool CondensedDMat::ContainsRow(int i)
{
    assert(mode == 0);
    return (index_map->count(i) > 0);
}

bool CondensedDMat::ContainsCol(int j)
{
    assert(mode == 1);
    return (index_map->count(j) > 0);
}

void CondensedDMat::RightDiagMult(VectorXd& v)
{
    assert(mode == 1);
    for (int j = 0; j < v.size(); ++j)
    {
        if (ContainsCol(j))
        {
            int c = (*index_map)[j];

            for (int r = 0; r < cond_mat->rows(); ++r)
            {
                (*cond_mat)(r, c) *= v(j);
            }
        }
    }
}

void CondensedDMat::RightDiagMult(Eigen::SparseVector<double>& v)
{
    assert(mode == 1);
    for (Eigen::SparseVector<double>::InnerIterator iter(v); iter; ++iter)
    {
        int r = iter.row();
        double val = iter.value();

        if (ContainsCol(r))
        {
            int c = (*index_map)[r];

            cond_mat->col(c) *= val;
        }
    }
}
