/* Ankur Parikh
   CondensedDMat.hpp
   
   takes a large dense matrix and stores it in a smaller 
   compressed (still dense) matrix where either rows that are all zero or columns that are all zero are removed
   the "mode" variable indicates whether rows/cols are being removed
   */

#pragma once

#include <iostream>
#include <fstream>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;

class CondensedDMat
{
private:

	map<int, int>* index_map; //maps old rows/cols to new rows/cols
	int mode; // 0 means rows are condensed, 1 means cols are condensed
	MatrixXd* cond_mat; // dense matrix

	friend class boost::serialization::access;

	template<typename Archive>
	void serialize(Archive& ar, const unsigned version) {
    
		ar & index_map;
		ar & mode; 
		ar & cond_mat;
	} 

public:
	
	CondensedDMat();
	~CondensedDMat();
	CondensedDMat(MatrixXd& m, map<int,int>& index_map, int mode);

	map<int,int>& GetIndexMap() { return *index_map; }
	MatrixXd& GetMat() { return *cond_mat; }

	double At(int i,int j);

	bool ContainsRow(int i);
	bool ContainsCol(int j);

	VectorXd Row(int i);
	VectorXd Col(int j);

	void RightDiagMult(VectorXd& v); //multiply by diag(v) on the right
	void RightDiagMult(Eigen::SparseVector<double>& v); // multiply by diag(v) on the right

};