/* Ankur Parikh
   CondensedSMat.hpp

   takes a large very sparse matrix and stores it in a smaller
   compressed (still sparse) matrix where rows/cols with all zeros are removed */

#pragma once

#include <iostream>
#include <fstream>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;

class CondensedSMat
{
private:

    map<int, int>* row_map; // maps old rows to new rows
    map<int, int>* col_map; // maps old cols to new cols
    SparseMatrix<double>* cond_mat; // compressed matrix

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & row_map;
        ar & col_map;
        ar & cond_mat;
    }

public:

    CondensedSMat();
    ~CondensedSMat();
    CondensedSMat(SparseMatrix<double>& big_mat);

    void Initialize(vector<Triplet<double> >& triplets, int V);

    SparseMatrix<double>& GetCondMat() { return *cond_mat; }

    map<int, int>& GetRowMap() { return *row_map; }
    map<int, int>& GetColMap() { return *col_map; }

    double Sum() { if (row_map->size() > 0) { return cond_mat->sum(); } else return 0.0; }

    void RightDiagMult(Eigen::SparseVector<double>& v); //multiply by diag(v) on the right
    bool ContainsRow(int i);
    bool ContainsCol(int j);
};