/* Ankur Parikh
   DiscountComputer.hpp

   see .hpp for class/function comments
   */


#include "DiscountComputer.hpp"



void DiscountComputer::ComputeDiscount(SparseMatrix<double>& disc_mat, VectorXd& gamma_vec,
    SparseMatrix<double>& bigrams,
    SparseMatrix<double>& p_bigram, double prev_P,
    vector<double> modD)
{
    disc_mat.resize(bigrams.rows(), bigrams.cols());
    double p_min = EigenHelper::Min(p_bigram);
    assert((p_min - 1.0) < 1e-10);
    disc_mat = p_bigram;

    for (int i = 0; i < bigrams.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(bigrams, i); it; ++it)
        {

            int r = it.row();
            int c = it.col();
            double pcount = it.value();
            double count = pow(pcount, 1.0 / prev_P);
            if (prev_P == 0.0)
                assert(0.0);

            if (abs(count - 1.0) < 1e-10)
            {
                disc_mat.coeffRef(r, c) *= modD[0];
            }
            else if (abs(count - 2.0) < 1e-10)
            {
                disc_mat.coeffRef(r, c) *= modD[1];
            }
            else if (abs(count - 3.0) < 1e-10 || count > 3.0)
            {
                disc_mat.coeffRef(r, c) *= modD[2];
            }
            else
            {
                assert(0);
            }
            if (bigrams.coeff(r, c) < disc_mat.coeff(r, c))
            {
                cout << pcount << "\t" << count << "\t" << bigrams.coeff(r, c) << "\t" << disc_mat.coeff(r, c) << "\t" << modD[0] << "\t" << modD[1] << "\t" << modD[2] << "\n";
                assert(disc_mat.coeff(r, c) <= bigrams.coeff(r, c));
            }
        }
    }

    EigenHelper::ColSums(disc_mat, gamma_vec);
}
