/* Ankur Parikh
   EM.hpp

   EM algorithm for learning the low rank approximation
   based on generalized KL divergence
   */

#include <complex>
#include <Eigen/Dense>
#include <Eigen/SVD>
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"

using namespace Eigen;

using namespace std;

class EM
{
public:

    static const int MAX_ITERS = 80;

    static const int NUM_CORES = 8;

    static const double PREC = 1e-5;

    // initialize the conditional probability tables randomly
    static void GenerateRandomCPT(VectorXd& v, int K);
    static void GenerateRandomCPT(MatrixXd& m, int V, int K);

    // EM where E step is run sequentially
    static void RunBigEM(MatrixXd& l_cpt, MatrixXd& r_cpt,
        SparseMatrix<double>& counts, int K, int num_iters);

    // EM where the E step is run in parallel
    static void RunBigParallelEM(MatrixXd& l_cpt, MatrixXd& r_cpt,
        SparseMatrix<double>& counts, int K, int num_iters, int num_cores);

    // can be ignored, same as previous function (TODO: needs to be refactored)
    static void RunBigParallelEM(MatrixXd& l_cpt, MatrixXd& r_cpt, MatrixXd& real_l_cpt,
        SparseMatrix<double>& counts, int K, int num_iters, int num_cores);

};
