/* Ankur Parikh
   EigenHelper.hpp

   additional helper functions to help process Eigen matrices that are not provided in Eigen
   */

# pragma once

#include <vector>
#include <complex>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "Tensor.hpp"

using namespace Eigen;


class EigenHelper
{
public:

    // pseudo inverse
    static MatrixXd pinv(MatrixXd& m1);

    // minimum element
    static double Min(SparseMatrix<double>& m);


    // print vectors/matrices/ sparse matrices
    static void PrintVec(VectorXd& v1);
    static void PrintMat(MatrixXd& m1);
    static void PrintSpMat(SparseMatrix<double> spmat);

    static void ConvertToBinary(SparseMatrix<double>& spmat, SparseMatrix<double>& b_spmat);

    // take elementwise power of matrix m to power P and store result in power_m
    static void Power(SparseMatrix<double>& m, SparseMatrix<double>& power_m, double P);

    // take elementwise inverse of dense/sparse vectors
    static void SparseInverse(Eigen::SparseVector<double>& v, Eigen::SparseVector<double>& inv_v);
    static void DenseInverse(VectorXd& p_denom, VectorXd& p_denom_inv);

    // row/column sums of sparse matrices
    static void RowSums(SparseMatrix<double>& m, VectorXd& rowsums);
    static void ColSums(SparseMatrix<double>& m, VectorXd& colsums);
    static void ColSums(SparseMatrix<int>& m, VectorXd& colsums);

    // row/column sums of sparse V by V matrices that are implicitly represented as vectors of Eigen triplets
    static void ColSums(vector<Triplet<double> >& triplets, map<int, double>& colsums, int V);
    static void RowSums(vector<Triplet<double> >& triplets, map<int, double>& rowsums, int V);

    // partition vector for EM parallel processing
    static void ParallelPartition(VectorXd& counts, vector<int>& start_indices, vector<int>& end_indices, int num_cores);
};
