/* Ankur Parikh
   IKNWrapper.hpp

   wrapper around Kneser Ney model for query probabilities efficiently,
   has a cache so that repeated queries can simply be looked up in the cache
   */

#pragma once

#include "IQuadKneserNey.hpp"
#include "IKneserNey.hpp"


using namespace std;
using namespace Eigen;


class IKNWrapper
{

private:

    SpecMoments* spec_mom;
    IKneserNey* kn; // 2nd /3rd order Kneser Ney model
    IQuadKneserNey* qkn; // 4th order Kneser Ney model (TODO: refactor so all in one class)

    int max_order; // maximum order of Kneser Ney model (2,3, or 4)


    tr1::unordered_map<iquadple, double, hash_Q>* quad_cache; // cache for quadgram queries
    tr1::unordered_map<itriple, double, hash_T>* triple_cache; // cache for trigram queries
    tr1::unordered_map<pair<int, int>, double, hash_P>* pair_cache; // cache for bigram queries

    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & spec_mom;
        ar & kn;
        ar & qkn;
        ar & max_order;
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        ar & spec_mom;
        ar & kn;
        ar & qkn;
        ar & max_order;

        quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
        triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
        pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
    }



    BOOST_SERIALIZATION_SPLIT_MEMBER()

public:

    IKNWrapper() {}
    IKNWrapper(SpecMoments* spec_mom, IKneserNey* kn, IQuadKneserNey* qkn, int max_order);
    IKNWrapper(SpecMoments* spec_mom, IKneserNey* kn, int max_order);
    ~IKNWrapper();

    // query using word strings
    double GetCondProb(const string& w, const string& wp, const string& wpp, const string& wppp);
    double GetCondProb(const string& w, const string& wp, const string& wpp);
    double GetCondProb(const string& w, const string& wp);
    double GetCondProb(const string& w);

    // query using word indices
    double GetCondProb(int i, int j, int k, int l);
    double GetCondProb(int i, int j, int k);
    double GetCondProb(int i, int j);
    double GetCondProb(int i);

    // test if out of vocabulary
    bool IsOOV(const string& w) { return (spec_mom->GetFinalID(w) == spec_mom->GetOOVIndex()); }
};

