/* Ankur Parikh
Interpolated 3rd and 2nd order Kneser Ney Models

TODO: Refactor 2,3,4th order models into a single framework

*/

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "IKneserNey.hpp"
#include "VectorPlus.hpp"
#include <vector>
#include <string>
#include <boost/algorithm/string.hpp>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <iostream>
#include <fstream>
#include "IOLibrary.hpp"

using namespace Eigen;
using namespace std;

//default constructor is used; 


using namespace std;


IKneserNey::IKneserNey(SpecMoments* spec_mom, IWittenBell* wb, bool m_flag)
{
    this->spec_mom = spec_mom;
    this->wb = wb;
    this->m_flag = m_flag;

    b_n1prev_counts = new map<pair<int, int>, int>();
    u_n1prev_counts = new  map<int, int>();

    VectorXd& unigrams = spec_mom->getUnigrams();
    denom_b_n1prev = new VectorXd(unigrams.size());
    denom_b_n1prev->fill(0);
    denom_u_n1prev = 0;

    denom_b_n1test = new VectorXd(unigrams.size());
    denom_b_n1test->fill(0);
    denom_n1test_sum = 0;

    D1 = 0;
    D2 = 0;
    D3plus = 0;
    max_d_limit = 9999999;

    bi_D1 = 0;
    bi_D2 = 0;
    bi_D3plus = 0;
    bi_max_d_limit = 9999999;

    bD1 = 0;
    bD2 = 0;
    bD3plus = 0;
    b_max_d_limit = max_d_limit;

    bidisc_map = new map<pair<int, int>, double>();
    unidisc_map = new map<int, double>();
    bi_unidisc_map = new map<int, double>();

    ComputeUCounts();
    ComputeBCounts();
    ComputeBiKneserDiscounts();
    ComputeBiUWeights();
    ComputeKneserDiscounts();
    ComputeBWeights();
    ComputeUWeights();

    //	ComputeDTrigrams();
}



IKneserNey::IKneserNey(SpecMoments* spec_mom, IWittenBell* wb, double D1, double D2, double D3plus,
    double bD1, double bD2, double bD3plus)
{
    this->spec_mom = spec_mom;
    this->wb = wb;
    this->m_flag = true;

    b_n1prev_counts = new map<pair<int, int>, int>();
    u_n1prev_counts = new  map<int, int>();

    VectorXd& unigrams = spec_mom->getUnigrams();
    denom_b_n1prev = new VectorXd(unigrams.size());
    denom_b_n1prev->fill(0);
    denom_u_n1prev = 0;

    denom_b_n1test = new VectorXd(unigrams.size());
    denom_b_n1test->fill(0);
    denom_n1test_sum = 0;

    this->D1 = D1;
    this->D2 = D2;
    this->D3plus = D3plus;
    max_d_limit = 9999999;

    bi_D1 = 0;
    bi_D2 = 0;
    bi_D3plus = 0;
    bi_max_d_limit = 9999999;

    this->bD1 = bD1;
    this->bD2 = bD2;
    this->bD3plus = bD3plus;
    b_max_d_limit = max_d_limit;

    bidisc_map = new map<pair<int, int>, double>();
    unidisc_map = new map<int, double>();
    bi_unidisc_map = new map<int, double>();

    ComputeUCounts();
    ComputeBCounts();
    ComputeBiKneserDiscounts();
    ComputeBiUWeights();
    ComputeBWeights();
    ComputeUWeights();

    //	ComputeDTrigrams();
}




void IKneserNey::ComputeUCounts()
{

    map<pair<int, int>, int>& b_counts = spec_mom->getBigramCounts();
    map<pair<int, int>, int>::iterator iter;

    map<int, set<int>*> u_n1prev_map;

    for (iter = b_counts.begin(); iter != b_counts.end(); ++iter)
    {
        pair<int, int> p = iter->first;

        // n1plus_set
        if (u_n1prev_map.count(p.first) == 0)
            u_n1prev_map[p.first] = new set<int>();

        u_n1prev_map[p.first]->insert(p.second);
    }


    map<int, set<int>*>::iterator prev_iter;

    for (prev_iter = u_n1prev_map.begin(); prev_iter != u_n1prev_map.end(); ++prev_iter)
    {
        int key = prev_iter->first;
        set<int>* val_set = prev_iter->second;
        int val = val_set->size();

        (*u_n1prev_counts)[key] = val;
        delete(val_set);

        denom_u_n1prev += val;
    }
}

void IKneserNey::ComputeBCounts()
{
    map<itriple, int>& t_counts = spec_mom->getTrigramCounts();
    map<pair<int, int>, int>& b_counts = spec_mom->getBigramCounts();

    map<itriple, int>::iterator iter;
    map<pair<int, int>, int>::iterator b_iter;
    map<pair<int, int>, set<int>*> b_n1prev_map;

    for (iter = t_counts.begin(); iter != t_counts.end(); ++iter)
    {
        itriple key = iter->first;

        pair<int, int> p(key.get<0>(), key.get<1>());

        // plus set
        if (b_n1prev_map.count(p) == 0)
            b_n1prev_map[p] = new set<int>();

        b_n1prev_map[p]->insert(key.get<2>());
    }

    // get the <s> pairs
    for (b_iter = b_counts.begin(); b_iter != b_counts.end(); ++b_iter)
    {
        pair<int, int> p = b_iter->first;

        if (p.second == 0)
        {
            (*b_n1prev_counts)[p] = b_iter->second;
            (*denom_b_n1prev)(p.second) += b_iter->second;
        }
    }

    // fill up plus and prev maps
    map<pair<int, int>, set<int>*>::iterator prev_iter;
    b_n1prev_mat = new SparseMatrix<double>(spec_mom->GetV(), spec_mom->GetV());
    vector<triplet> triplet_vec;
    for (prev_iter = b_n1prev_map.begin(); prev_iter != b_n1prev_map.end(); ++prev_iter)
    {
        pair<int, int> key = prev_iter->first;
        set<int>* val_set = prev_iter->second;
        int val = val_set->size();
        if (val == 0)
            assert(0);

        (*b_n1prev_counts)[key] = val;
        delete(val_set);

        triplet_vec.push_back(triplet(key.first, key.second, val));
        (*denom_b_n1prev)(key.second) += val;
        (*denom_b_n1test)(key.first) += 1.0;
    }

    b_n1prev_mat->setFromTriplets(triplet_vec.begin(), triplet_vec.end());
    ComputeTriUniKneserDiscounts();
    denom_n1test_sum = denom_b_n1test->sum();



}

void IKneserNey::ComputeBWeights()
{
    map<pair<int, int>, int> bimap1;
    map<pair<int, int>, int> bimap2;
    map<pair<int, int>, int> bimap3plus;

    map<itriple, int>::iterator iter;

    map<itriple, int>& trigram_counts = spec_mom->getTrigramCounts();

    for (iter = trigram_counts.begin(); iter != trigram_counts.end(); ++iter)
    {
        itriple t = iter->first;
        pair<int, int> p(t.get<1>(), t.get<2>());
        int count = iter->second;

        if (count == 1)
        {
            if (bimap1.count(p) == 0)
                bimap1[p] = 1;
            else
                ++(bimap1[p]);

        }
        else if (count == 2)
        {
            if (bimap2.count(p) == 0)
                bimap2[p] = 1;
            else
                ++(bimap2[p]);

        }
        else if (count >= 3 && count <= max_d_limit)
        {
            if (bimap3plus.count(p) == 0)
                bimap3plus[p] = 1;
            else
                ++(bimap3plus[p]);

        }
        else
        {
            assert(count > max_d_limit);
        }
    }

    // iterate through all the bigrams

    map<pair<int, int>, int>::iterator iter2;
    map<pair<int, int>, int>& bh_counts = spec_mom->GetBHCountMap();
    for (iter2 = bh_counts.begin(); iter2 != bh_counts.end(); ++iter2)
    {
        pair<int, int> p = iter2->first;
        int denom = iter2->second;

        double numer = 0;
        if (bimap1.count(p) > 0)
        {
            numer += D1 * bimap1[p];
        }
        if (bimap2.count(p) > 0)
        {
            numer += D2 * bimap2[p];
        }
        if (bimap3plus.count(p) > 0)
        {
            numer += D3plus * bimap3plus[p];
        }
        assert(numer > 0);


        (*bidisc_map)[p] = numer / denom;
    }
}


void IKneserNey::ComputeUWeights()
{
    map<int, int> unimap1;
    map<int, int> unimap2;
    map<int, int> unimap3plus;

    map<pair<int, int>, int>::iterator iter;

    for (iter = b_n1prev_counts->begin(); iter != b_n1prev_counts->end(); ++iter)
    {
        pair<int, int> p = iter->first;
        int count = iter->second;

        int ukey = p.second;

        if (count == 1)
        {
            if (unimap1.count(ukey) == 0)
                unimap1[ukey] = 1;
            else
                ++(unimap1[ukey]);

        }
        else if (count == 2)
        {
            if (unimap2.count(ukey) == 0)
                unimap2[ukey] = 1;
            else
                ++(unimap2[ukey]);

        }
        else if (count >= 3 && count <= max_d_limit)
        {
            if (unimap3plus.count(ukey) == 0)
                unimap3plus[ukey] = 1;
            else
                ++(unimap3plus[ukey]);

        }
        else
        {
            assert(count > max_d_limit);
        }
    }

    // iterate through all the bigrams
    for (int i = 0; i < denom_b_n1prev->size(); ++i)
    {
        int key = i;
        double denom = (*denom_b_n1prev)(i);
        if (denom < 1e-20)
            continue;
        //	cout << "denom: " + ToString(denom) + "\n";
        double numer = 0;
        if (unimap1.count(key) > 0)
        {
            numer += bD1 * unimap1[key];
        }
        if (unimap2.count(key) > 0)
        {
            numer += bD2 * unimap2[key];
        }
        if (unimap3plus.count(key) > 0)
        {
            numer += bD3plus * unimap3plus[key];
        }
        assert(numer > 0);

        (*unidisc_map)[key] = numer / denom;
    }
}

void IKneserNey::ComputeBiUWeights()
{
    map<int, int> unimap1;
    map<int, int> unimap2;
    map<int, int> unimap3plus;

    map<pair<int, int>, int>::iterator iter;

    map<pair<int, int>, int>& bigram_counts = spec_mom->getBigramCounts();

    for (iter = bigram_counts.begin(); iter != bigram_counts.end(); ++iter)
    {
        pair<int, int> p = iter->first;
        int count = iter->second;

        int ukey = p.second;

        if (count == 1)
        {
            if (unimap1.count(ukey) == 0)
                unimap1[ukey] = 1;
            else
                ++(unimap1[ukey]);

        }
        else if (count == 2)
        {
            if (unimap2.count(ukey) == 0)
                unimap2[ukey] = 1;
            else
                ++(unimap2[ukey]);

        }
        else if (count >= 3 && count <= max_d_limit)
        {
            if (unimap3plus.count(ukey) == 0)
                unimap3plus[ukey] = 1;
            else
                ++(unimap3plus[ukey]);

        }
        else
        {
            assert(count > max_d_limit);
        }
    }

    // iterate through all the denom_b_n1prev

    map<int, int>& uh_map = spec_mom->GetUHCountMap();
    map<int, int>::iterator u_iter;
    for (u_iter = uh_map.begin(); u_iter != uh_map.end(); ++u_iter)
    {
        int key = u_iter->first;
        int denom = u_iter->second;
        if (denom == 0)
            assert(0);
        //	cout << "denom: " + ToString(denom) + "\n";
        double numer = 0;
        if (unimap1.count(key) > 0)
        {
            numer += bi_D1 * unimap1[key];
        }
        if (unimap2.count(key) > 0)
        {
            numer += bi_D2 * unimap2[key];
        }
        if (unimap3plus.count(key) > 0)
        {
            numer += bi_D3plus * unimap3plus[key];
        }
        assert(numer > 0);

        (*bi_unidisc_map)[key] = numer / denom;


    }
}

double IKneserNey::GetTriDiscountedUnigramProb(int i)
{

    double numer = (*denom_b_n1test)(i);
    return (numer + 0.0) / (denom_n1test_sum + 0.0);
}


double IKneserNey::GetDiscountedUnigramProb(int i)
{

    double numer = GetN1Prev(i);

    return (numer + 0.0) / (denom_u_n1prev + 0.0);

}


void IKneserNey::ComputeTriUniKneserDiscounts()
{
    double N1 = 0;
    double N2 = 0;

    int non_zero = 0;
    for (int i = 0; i < denom_b_n1test->size(); ++i)
    {
        if (denom_b_n1test->coeff(i) == 1.0)
            N1++;
        else if (denom_b_n1test->coeff(i) == 2.0)
            N2++;

        if (denom_b_n1test->coeff(i) > 0)
        {
            assert(denom_b_n1test->coeff(i) >= 1);
            non_zero++;
        }
    }

    double uni_Y = N1 / (N1 + 2 * N2);
    uni_D1 = 1 - 2 * uni_Y*(N2 / N1);

    double sum = denom_b_n1test->sum();
    double gamma = non_zero * uni_D1 / denom_b_n1test->size();

    assert(gamma <= 1);

    for (int i = 0; i < denom_b_n1test->size(); ++i)
    {
        double val = denom_b_n1test->coeff(i);

        if (val > 0)
        {
            assert(val >= 1);
            denom_b_n1test->coeffRef(i) -= uni_D1;
        }
        denom_b_n1test->coeffRef(i) += gamma;
    }
    assert(abs(denom_b_n1test->sum() - sum) < 1e-3);
}


void IKneserNey::ComputeBiKneserDiscounts()
{
    map<int, int> bi_cc;

    map<pair<int, int>, int>::iterator bi_iter;
    map<pair<int, int>, int>& bigram_counts = spec_mom->getBigramCounts();

    for (bi_iter = bigram_counts.begin(); bi_iter != bigram_counts.end(); ++bi_iter)
    {
        int count = bi_iter->second;

        if (bi_cc.count(count) > 0)
            ++(bi_cc[count]);
        else
            bi_cc[count] = 1;
    }

    //trigram discount
    double bi_N1 = bi_cc[1];
    double bi_N2 = bi_cc[2];
    double bi_N3 = bi_cc[3];
    double bi_N4 = bi_cc[4];

    double bi_Y = bi_N1 / (bi_N1 + 2 * bi_N2);
    bi_D1 = 1 - 2 * bi_Y*(bi_N2 / bi_N1);
    bi_D2 = 2 - 3 * bi_Y*(bi_N3 / bi_N2);
    bi_D3plus = 3 - 4 * bi_Y*(bi_N4 / bi_N3);

    if (!m_flag)
    {
        bi_D2 = bi_D1;
        bi_D3plus = bi_D1;
    }


    //	assert(bi_D1 > 0 && bi_D1 <= 1);
    //	assert(bi_D2 > 0 && bi_D2 <= 2);
    //	assert(bi_D3plus > 0 && bi_D3plus <= 3);
}





// discount trigram counts
void IKneserNey::ComputeKneserDiscounts()
{
    map<int, int> tri_cc;
    map<int, int> bi_cc;
    map<int, int> uni_cc;

    map<int, int>::iterator uni_iter;
    map<pair<int, int>, int>::iterator bi_iter;
    map<itriple, int>::iterator tri_iter;

    map<int, int>& unigram_counts = spec_mom->getUnigramCounts();
    map<pair<int, int>, int>& bigram_counts = spec_mom->getBigramCounts();
    map<itriple, int>& trigram_counts = spec_mom->getTrigramCounts();

    for (tri_iter = trigram_counts.begin(); tri_iter != trigram_counts.end(); ++tri_iter)
    {
        int count = tri_iter->second;

        if (tri_cc.count(count) > 0)
            ++(tri_cc[count]);
        else
            tri_cc[count] = 1;
    }

    for (bi_iter = b_n1prev_counts->begin(); bi_iter != b_n1prev_counts->end(); ++bi_iter)
    {
        int count = bi_iter->second;

        if (bi_cc.count(count) > 0)
            ++(bi_cc[count]);
        else
            bi_cc[count] = 1;
    }

    //trigram discount
    double N1 = tri_cc[1];
    double N2 = tri_cc[2];
    double N3 = tri_cc[3];
    double N4 = tri_cc[4];

    double Y = N1 / (N1 + 2 * N2);
    D1 = 1 - 2 * Y*(N2 / N1);
    D2 = 2 - 3 * Y*(N3 / N2);
    D3plus = 3 - 4 * Y*(N4 / N3);

    if (!m_flag)
    {
        D2 = D1;
        D3plus = D1;
    }

    //	assert(D1 > 0 && D1 < 1);
    //	assert(D2 > 0 && D2 < 2);
    //	assert(D3plus > 0 && D3plus < 3);

    //bigram discounts
    double bN1 = bi_cc[1];
    double bN2 = bi_cc[2];
    double bN3 = bi_cc[3];
    double bN4 = bi_cc[4];

    double bY = bN1 / (bN1 + 2 * bN2);
    bD1 = 1 - 2 * bY*(bN2 / bN1);
    bD2 = 2 - 3 * bY*(bN3 / bN2);
    bD3plus = 3 - 4 * bY*(bN4 / bN3);

    if (!m_flag)
    {
        bD2 = bD1;
        bD3plus = bD1;
    }

    //	assert(bD1 > 0 && bD1 < 1);
    //	assert(bD2 > 0 && bD2 < 2);
    //	assert(bD3plus > 0 && bD3plus < 3);
}



double IKneserNey::GetDiscountedTrigramProb(int i, int j, int k)
{
    itriple v(i, j, k);
    pair<int, int> p(j, k);

    map<itriple, int>& trigram_counts = spec_mom->getTrigramCounts();

    if (trigram_counts.count(v) == 0)
        return 0;
    else
    {
        int tc = trigram_counts[v];
        int bc = spec_mom->GetBHCount(p);

        double d_tc = tc;
        if (tc == 1)
        {
            d_tc -= D1;
        }
        else if (tc == 2)
        {
            d_tc -= D2;
        }
        else if (tc >= 3 && tc <= max_d_limit)
        {
            d_tc -= D3plus;
        }

        double val = d_tc / bc;
        return val;
    }
}


double IKneserNey::GetBiDiscBigramProb(int i, int j)
{
    pair<int, int> p(i, j);

    map<pair<int, int>, int>& bigram_counts = spec_mom->getBigramCounts();

    if (bigram_counts.count(p) == 0)
        return 0;
    else
    {
        int bc = bigram_counts[p];
        int uc = spec_mom->GetUHCount(j);

        double d_bc = bc;
        if (bc == 1)
        {
            d_bc -= bi_D1;
        }
        else if (bc == 2)
        {
            //	d_bc -= bi_D1;
            d_bc -= bi_D2;
        }
        else if (bc >= 3 && bc <= max_d_limit)
        {
            //	d_bc -= bi_D1;
            d_bc -= bi_D3plus;
        }

        double val = d_bc / uc;
        return val;
    }
}

double IKneserNey::GetNonDiscBigramProb(int i, int j)
{
    double n1prev_c = (double)GetN1Prev(i, j);
    double numer = 0.0;
    if (n1prev_c == 1)
    {
        numer = n1prev_c - bD1;
    }
    else if (n1prev_c == 2)
    {
        numer = n1prev_c - bD2;
    }
    else if (n1prev_c >= 3 && n1prev_c <= max_d_limit)
    {
        numer = n1prev_c - bD3plus;
    }

    double denom = (*denom_b_n1prev)(j);
    double val = n1prev_c / denom;
    return val;
}


double IKneserNey::GetDiscountedBigramProb(int i, int j)
{
    double n1prev_c = (double)GetN1Prev(i, j);
    double numer = 0.0;
    if (n1prev_c == 1)
    {
        numer = n1prev_c - bD1;
    }
    else if (n1prev_c == 2)
    {
        numer = n1prev_c - bD2;
    }
    else if (n1prev_c >= 3 && n1prev_c <= max_d_limit)
    {
        numer = n1prev_c - bD3plus;
    }

    double denom = (*denom_b_n1prev)(j);
    double val = numer / denom;
    return val;
}



double IKneserNey::GetN1Prev(int i)
{
    if (u_n1prev_counts->count(i) > 0)
    {
        return (*u_n1prev_counts)[i];
    }
    else
        return 0;
}

double IKneserNey::GetN1Prev(int i, int j)
{
    pair<int, int> p(i, j);

    if (b_n1prev_counts->count(p) > 0)
        return (*b_n1prev_counts)[p];
    else
        return 0;
}
