/* Ankur Parikh
   IOLibrary.hpp

   Helper functions for input/output from files
   */

# pragma once


#include<string>
#include <Eigen/Dense>
#include <sstream>
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/algorithm/string.hpp>
#include "SpecMoments.hpp"

using namespace std;
using namespace Eigen;


// read matrix from file fn where n_cols is number of columns 
template <class T>
void ReadMatrix(Matrix<T, Dynamic, Dynamic>& m, string fn, int n_cols)
{
    ifstream m_file;
    m_file.open(fn.c_str());
    string line;

    vector<T> vals;
    int n_rows = 0;

    while (getline(m_file, line))
    {
        stringstream ss(line);
        string delim_line;

        for (int i = 0; i < n_cols; ++i)
        {
            getline(ss, delim_line, '\t');
            double val = atof(delim_line.c_str());
            vals.push_back((T)val);
        }
        n_rows++;
    }
    m_file.close();

    if (n_rows == 0)
        assert(0);

    int index = 0;
    m.resize(n_rows, n_cols);

    for (int i = 0; i < n_rows; ++i)
    {
        for (int j = 0; j < n_cols; ++j)
        {
            m(i, j) = vals[index++];
        }
    }
}

// converts t of type T to string
template <class T>
string ToString(T t)
{
    stringstream ss;
    ss << t;
    return ss.str();
}

// write matrix to file named fn
template <class T>
void WriteMatrix(Matrix<T, Dynamic, Dynamic>& m, string fn)
{
    ofstream m_file;
    m_file.open(fn.c_str());
    m_file << m;
    m_file.close();
}

// read test_fn into a vector of vectors 
static void GetBigTestMatrix(vector< vector<int>* >& test_data, string test_fn, SpecMoments& m)
{
    ifstream test_file;
    test_file.open(test_fn.c_str());
    string line;

    // basically we first need to count and find max columns
    int max_col = 0;
    int rows = 0;
    int wc = 0;
    while (getline(test_file, line))
    {
        boost::trim(line);

        vector<string> strs;
        boost::split(strs, line, boost::is_any_of(" "));
        if (strs.size() == 0)
            continue;
        test_data.push_back(new vector<int>());
        test_data.at(test_data.size() - 1)->reserve(strs.size());
        rows++;

        for (int c = 0; c < strs.size(); ++c)
        {
            int val = m.GetID(strs[c]);
            if (val != std::numeric_limits<int>::max())
                wc++;
            test_data.at(test_data.size() - 1)->push_back(val);
        }
    }
    test_file.close();

    if (rows == 0)
        assert(0);
}


// read test_fn into a matrix
static void GetIOTestMatrix(MatrixXi& test_data, string test_fn, SpecMoments& m)
{
    ifstream test_file;
    test_file.open(test_fn.c_str());
    string line;

    // basically we first need to count and find max columns
    int max_col = 0;
    int rows = 0;
    int wc = 0;
    while (getline(test_file, line))
    {
        boost::trim(line);
        rows++;
        vector<string> strs;
        boost::split(strs, line, boost::is_any_of(" "));

        if (strs.size() > max_col)
            max_col = strs.size();

    }
    test_file.close();

    if (rows == 0 || max_col == 0)
        assert(0);

    test_data.resize(rows, max_col);
    test_data.fill(-1);
    test_file.open(test_fn.c_str());
    int r_index = 0;
    while (getline(test_file, line))
    {
        boost::trim(line);

        vector<string> strs;
        boost::split(strs, line, boost::is_any_of(" "));

        for (int c = 0; c < strs.size(); ++c)
        {
            int val = m.GetID(strs[c]);
            if (val != std::numeric_limits<int>::max())
                wc++;
            test_data(r_index, c) = val;
        }

        r_index++;
    }
}
// read matrix from fn quickly since number of rows (n_rows) and number of columns 
// (n_cols) is specified
static void ReadFastMatrix(MatrixXd& m, string fn, int n_rows, int n_cols)
{
    ifstream m_file;
    m_file.open(fn.c_str());
    string line;
    m.resize(n_rows, n_cols);
    int index = 0;
    while (getline(m_file, line))
    {
        stringstream ss(line);
        string delim_line;
        vector<string> strs;
        boost::split(strs, line, boost::is_any_of("\t"));
        int c = 0;
        for (int i = 0; i < strs.size(); ++i)
        {
            if (strs[i] == "")
                continue;

            double val = atof(strs[i].c_str());

            m(index, c) = val;
            c++;

        }
        assert(c == n_cols);
        index++;
    }
    assert(index == n_rows);
    m_file.close();
}