/* Ankur Parikh
    IQuadKneserNey.cpp

    see .hpp for class/function comments
    */

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "IQuadKneserNey.hpp"
#include "VectorPlus.hpp"
#include <vector>
#include <string>
#include <boost/algorithm/string.hpp>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <iostream>
#include <fstream>
#include "IOLibrary.hpp"

using namespace Eigen;
using namespace std;

//default constructor is used; 


using namespace std;


IQuadKneserNey::IQuadKneserNey(SpecMoments* spec_mom, IWittenBell* wb, bool m_flag)
{
    this->spec_mom = spec_mom;
    this->wb = wb;
    this->m_flag = m_flag;

    t_n1prev_counts = new map<itriple, int>();
    b_n1prev_counts = new map<pair<int, int>, int>();
    u_n1prev_counts = new  map<int, int>();

    VectorXd& unigrams = spec_mom->getUnigrams();

    //	denom_t_n1prev = new SparseMatrix<double>(unigrams.size(), unigrams.size());
    denom_t_n1prev = new SparseMatrix<double>(unigrams.size(), unigrams.size());
    //	denom_t_n1prev->fill(0.0);

    denom_b_n1prev = new VectorXd(unigrams.size());
    denom_b_n1prev->fill(0);
    denom_u_n1prev = 0;

    denom_b_n1test = new VectorXd(unigrams.size());
    denom_b_n1test->fill(0);
    denom_n1test_sum = 0;

    qD1 = 0;
    qD2 = 0;
    qD3plus = 0;
    q_max_d_limit = 1e15;

    tD1 = 0;
    tD2 = 0;
    tD3plus = 0;
    t_max_d_limit = 1e15;

    bD1 = 0;
    bD2 = 0;
    bD3plus = 0;
    b_max_d_limit = 1e15;

    tridisc_map = new map<itriple, double>();
    bidisc_map = new map<pair<int, int>, double>();
    unidisc_map = new map<int, double>();

    ComputeUCounts();
    ComputeBCounts();
    ComputeTCounts();
    ComputeBiKneserDiscounts();
    ComputeTriKneserDiscounts();
    ComputeQuadKneserDiscounts();
    ComputeTWeights();
    ComputeBWeights();
    ComputeUWeights();
}

void IQuadKneserNey::ComputeUCounts()
{

    map<pair<int, int>, int>& b_counts = spec_mom->getBigramCounts();
    map<pair<int, int>, int>::iterator iter;

    map<int, set<int>*> u_n1prev_map;

    for (iter = b_counts.begin(); iter != b_counts.end(); ++iter)
    {
        pair<int, int> p = iter->first;

        // n1plus_set
        if (u_n1prev_map.count(p.first) == 0)
            u_n1prev_map[p.first] = new set<int>();

        u_n1prev_map[p.first]->insert(p.second);
    }


    map<int, set<int>*>::iterator prev_iter;

    for (prev_iter = u_n1prev_map.begin(); prev_iter != u_n1prev_map.end(); ++prev_iter)
    {
        int key = prev_iter->first;
        set<int>* val_set = prev_iter->second;
        int val = val_set->size();

        (*u_n1prev_counts)[key] = val;
        delete(val_set);

        denom_u_n1prev += val;
    }
}

void IQuadKneserNey::ComputeBCounts()
{
    map<itriple, int>& t_counts = spec_mom->getTrigramCounts();
    map<pair<int, int>, int>& b_counts = spec_mom->getBigramCounts();

    map<itriple, int>::iterator iter;
    map<pair<int, int>, int>::iterator b_iter;
    map<pair<int, int>, set<int>*> b_n1prev_map;

    for (iter = t_counts.begin(); iter != t_counts.end(); ++iter)
    {
        itriple key = iter->first;

        pair<int, int> p(key.get<0>(), key.get<1>());

        // plus set
        if (b_n1prev_map.count(p) == 0)
            b_n1prev_map[p] = new set<int>();

        b_n1prev_map[p]->insert(key.get<2>());


    }

    // get the <s> pairs
    for (b_iter = b_counts.begin(); b_iter != b_counts.end(); ++b_iter)
    {
        pair<int, int> p = b_iter->first;

        if (p.second == 0)
        {
            (*b_n1prev_counts)[p] = b_iter->second;

            (*denom_b_n1prev)(p.second) += b_iter->second;
        }
    }

    // fill up plus and prev maps
    map<pair<int, int>, set<int>*>::iterator prev_iter;
    b_n1prev_mat = new SparseMatrix<double>(spec_mom->GetV(), spec_mom->GetV());
    vector<triplet> triplet_vec;
    for (prev_iter = b_n1prev_map.begin(); prev_iter != b_n1prev_map.end(); ++prev_iter)
    {
        pair<int, int> key = prev_iter->first;
        set<int>* val_set = prev_iter->second;
        int val = val_set->size();
        if (val == 0)
            assert(0);

        (*b_n1prev_counts)[key] = val;
        delete(val_set);

        triplet_vec.push_back(triplet(key.first, key.second, val));
        (*denom_b_n1prev)(key.second) += val;
        (*denom_b_n1test)(key.first) += 1.0;
    }

    b_n1prev_mat->setFromTriplets(triplet_vec.begin(), triplet_vec.end());
    ComputeTriUniKneserDiscounts();
    denom_n1test_sum = denom_b_n1test->sum();
}



void IQuadKneserNey::ComputeTCounts()
{
    map<iquadple, int>& q_counts = spec_mom->getQuadgramCounts();
    map<itriple, int>& t_counts = spec_mom->getTrigramCounts();

    map<pair<int, int>, double> denom_map;

    map<iquadple, int>::iterator qiter;
    map<itriple, int>::iterator titer;
    map<itriple, set<int>*> t_n1prev_map;

    for (qiter = q_counts.begin(); qiter != q_counts.end(); ++qiter)
    {
        iquadple key = qiter->first;

        itriple t(key.get<0>(), key.get<1>(), key.get<2>());

        // plus set
        if (t_n1prev_map.count(t) == 0)
            t_n1prev_map[t] = new set<int>();

        t_n1prev_map[t]->insert(key.get<3>());


    }

    // get the <s> pairs
    for (titer = t_counts.begin(); titer != t_counts.end(); ++titer)
    {
        itriple t = titer->first;

        if (t.get<2>() == 0)
        {
            (*t_n1prev_counts)[t] = titer->second;
            pair<int, int> p(t.get<1>(), t.get<2>());
            if (denom_map.count(p) == 0)
                denom_map[p] = 0;
            denom_map[p] += titer->second;
        }
    }

    // fill up plus and prev maps
    map<itriple, set<int>*>::iterator prev_iter;

    for (prev_iter = t_n1prev_map.begin(); prev_iter != t_n1prev_map.end(); ++prev_iter)
    {
        itriple key = prev_iter->first;
        set<int>* val_set = prev_iter->second;
        int val = val_set->size();
        if (val == 0)
            assert(0);

        (*t_n1prev_counts)[key] = val;
        delete(val_set);

        pair<int, int> p(key.get<1>(), key.get<2>());
        if (denom_map.count(p) == 0)
            denom_map[p] = 0;

        denom_map[p] += val;
        //  (*denom_t_n1prev)(key[1], key[2]) += val;
    }



    vector<triplet> denom_triplets;
    map<pair<int, int>, double>::iterator diter;
    for (diter = denom_map.begin(); diter != denom_map.end(); ++diter)
    {
        pair<int, int> p = diter->first;
        double total = diter->second;

        denom_triplets.push_back(triplet(p.first, p.second, total));
    }
    denom_t_n1prev->setFromTriplets(denom_triplets.begin(), denom_triplets.end());
}


void IQuadKneserNey::ComputeTWeights()
{
    map<itriple, int> trimap1;
    map<itriple, int> trimap2;
    map<itriple, int> trimap3plus;

    map<iquadple, int>::iterator iter;

    map<iquadple, int>& quadgram_counts = spec_mom->getQuadgramCounts();

    for (iter = quadgram_counts.begin(); iter != quadgram_counts.end(); ++iter)
    {
        iquadple qvec = iter->first;
        itriple t(qvec.get<1>(), qvec.get<2>(), qvec.get<3>());

        int count = iter->second;

        if (count == 1)
        {
            if (trimap1.count(t) == 0)
                trimap1[t] = 1;
            else
                ++(trimap1[t]);

        }
        else if (count == 2)
        {
            if (trimap2.count(t) == 0)
                trimap2[t] = 1;
            else
                ++(trimap2[t]);

        }
        else if (count >= 3 && count <= t_max_d_limit)
        {
            if (trimap3plus.count(t) == 0)
                trimap3plus[t] = 1;
            else
                ++(trimap3plus[t]);

        }
        else
        {
            assert(count > t_max_d_limit);
        }
    }

    map<itriple, int>::iterator iter2;
    map<itriple, int>& th_counts = spec_mom->GetTHCountMap();
    for (iter2 = th_counts.begin(); iter2 != th_counts.end(); ++iter2)
    {
        itriple v = iter2->first;
        int denom = iter2->second;

        double numer = 0;
        if (trimap1.count(v) > 0)
        {
            numer += qD1 * trimap1[v];
        }
        if (trimap2.count(v) > 0)
        {
            numer += qD2 * trimap2[v];
        }
        if (trimap3plus.count(v) > 0)
        {
            numer += qD3plus * trimap3plus[v];
        }
        assert(numer > 0);


        (*tridisc_map)[v] = numer / denom;
    }
}


void IQuadKneserNey::ComputeBWeights()
{
    map<pair<int, int>, int> bimap1;
    map<pair<int, int>, int> bimap2;
    map<pair<int, int>, int> bimap3plus;

    map<itriple, int>::iterator iter;

    for (iter = t_n1prev_counts->begin(); iter != t_n1prev_counts->end(); ++iter)
    {
        itriple t = iter->first;
        pair<int, int> p(t.get<1>(), t.get<2>());
        int count = iter->second;

        if (count == 1)
        {
            if (bimap1.count(p) == 0)
                bimap1[p] = 1;
            else
                ++(bimap1[p]);

        }
        else if (count == 2)
        {
            if (bimap2.count(p) == 0)
                bimap2[p] = 1;
            else
                ++(bimap2[p]);

        }
        else if (count >= 3 && count <= t_max_d_limit)
        {
            if (bimap3plus.count(p) == 0)
                bimap3plus[p] = 1;
            else
                ++(bimap3plus[p]);

        }
        else
        {
            assert(count > t_max_d_limit);
        }
    }


    int V = spec_mom->GetV();


    for (int i = 0; i < denom_t_n1prev->outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(*denom_t_n1prev, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double denom = it.value();

            pair<int, int> p(r, c);
            if (denom == 0.0)
                continue;

            double numer = 0;
            if (bimap1.count(p) > 0)
            {
                numer += tD1 * bimap1[p];
            }
            if (bimap2.count(p) > 0)
            {
                numer += tD2 * bimap2[p];
            }
            if (bimap3plus.count(p) > 0)
            {
                numer += tD3plus * bimap3plus[p];
            }
            assert(numer > 0);

            (*bidisc_map)[p] = numer / denom;
        }
    }
}


void IQuadKneserNey::ComputeUWeights()
{
    map<int, int> unimap1;
    map<int, int> unimap2;
    map<int, int> unimap3plus;

    map<pair<int, int>, int>::iterator iter;

    for (iter = b_n1prev_counts->begin(); iter != b_n1prev_counts->end(); ++iter)
    {
        pair<int, int> p = iter->first;
        int count = iter->second;

        int ukey = p.second;

        if (count == 1)
        {
            if (unimap1.count(ukey) == 0)
                unimap1[ukey] = 1;
            else
                ++(unimap1[ukey]);

        }
        else if (count == 2)
        {
            if (unimap2.count(ukey) == 0)
                unimap2[ukey] = 1;
            else
                ++(unimap2[ukey]);

        }
        else if (count >= 3 && count <= b_max_d_limit)
        {
            if (unimap3plus.count(ukey) == 0)
                unimap3plus[ukey] = 1;
            else
                ++(unimap3plus[ukey]);

        }
        else
        {
            assert(count > b_max_d_limit);
        }
    }

    // iterate through all the bigrams
    for (int i = 0; i < denom_b_n1prev->size(); ++i)
    {
        int key = i;
        double denom = (*denom_b_n1prev)(i);
        if (denom < 1e-20)
            continue;
        double numer = 0;
        if (unimap1.count(key) > 0)
        {
            numer += bD1 * unimap1[key];
        }
        if (unimap2.count(key) > 0)
        {
            numer += bD2 * unimap2[key];
        }
        if (unimap3plus.count(key) > 0)
        {
            numer += bD3plus * unimap3plus[key];
        }
        assert(numer > 0);

        (*unidisc_map)[key] = numer / denom;
    }
}



double IQuadKneserNey::GetTriDiscountedUnigramProb(int i)
{

    double numer = (*denom_b_n1test)(i);
    return (numer + 0.0) / (denom_n1test_sum + 0.0);
}


double IQuadKneserNey::GetDiscountedUnigramProb(int i)
{

    double numer = GetN1Prev(i);

    return (numer + 0.0) / (denom_u_n1prev + 0.0);

}


void IQuadKneserNey::ComputeTriUniKneserDiscounts()
{
    double N1 = 0;
    double N2 = 0;

    int non_zero = 0;
    for (int i = 0; i < denom_b_n1test->size(); ++i)
    {
        if (denom_b_n1test->coeff(i) == 1.0)
            N1++;
        else if (denom_b_n1test->coeff(i) == 2.0)
            N2++;

        if (denom_b_n1test->coeff(i) > 0)
        {
            assert(denom_b_n1test->coeff(i) >= 1);
            non_zero++;
        }
    }

    double uni_Y = N1 / (N1 + 2 * N2);
    double uni_D1 = 1 - 2 * uni_Y*(N2 / N1);

    double sum = denom_b_n1test->sum();
    double gamma = non_zero * uni_D1 / denom_b_n1test->size();

    assert(gamma <= 1);

    for (int i = 0; i < denom_b_n1test->size(); ++i)
    {
        double val = denom_b_n1test->coeff(i);

        if (val > 0)
        {
            assert(val >= 1);
            denom_b_n1test->coeffRef(i) -= uni_D1;
        }
        denom_b_n1test->coeffRef(i) += gamma;
    }
    assert(abs(denom_b_n1test->sum() - sum) < 1e-3);
}


void IQuadKneserNey::ComputeQuadKneserDiscounts()
{
    map<int, int> quad_cc;

    map<iquadple, int>::iterator quad_iter;

    map<iquadple, int>& quadgram_counts = spec_mom->getQuadgramCounts();

    for (quad_iter = quadgram_counts.begin(); quad_iter != quadgram_counts.end(); ++quad_iter)
    {
        int count = quad_iter->second;

        if (quad_cc.count(count) > 0)
            ++(quad_cc[count]);
        else
            quad_cc[count] = 1;
    }


    //quadgram discount
    double N1 = quad_cc[1];
    double N2 = quad_cc[2];
    double N3 = quad_cc[3];
    double N4 = quad_cc[4];


    double Y = N1 / (N1 + 2 * N2);
    qD1 = 1 - 2 * Y*(N2 / N1);
    qD2 = 2 - 3 * Y*(N3 / N2);
    qD3plus = 3 - 4 * Y*(N4 / N3);

    if (!m_flag)
    {
        qD2 = qD1;
        qD3plus = qD1;
    }
}

void IQuadKneserNey::ComputeTriKneserDiscounts()
{
    map<int, int> tri_cc;
    map<itriple, int>::iterator tri_iter;

    for (tri_iter = t_n1prev_counts->begin(); tri_iter != t_n1prev_counts->end(); ++tri_iter)
    {
        int count = tri_iter->second;

        if (tri_cc.count(count) > 0)
            ++(tri_cc[count]);
        else
            tri_cc[count] = 1;
    }

    //trigram discounts
    double N1 = tri_cc[1];
    double N2 = tri_cc[2];
    double N3 = tri_cc[3];
    double N4 = tri_cc[4];

    double Y = N1 / (N1 + 2 * N2);
    tD1 = 1 - 2 * Y*(N2 / N1);
    tD2 = 2 - 3 * Y*(N3 / N2);
    tD3plus = 3 - 4 * Y*(N4 / N3);

    if (!m_flag)
    {
        tD2 = tD1;
        tD3plus = tD1;
    }
}


// discount trigram counts
void IQuadKneserNey::ComputeBiKneserDiscounts()
{
    map<int, int> bi_cc;
    map<pair<int, int>, int>::iterator bi_iter;

    for (bi_iter = b_n1prev_counts->begin(); bi_iter != b_n1prev_counts->end(); ++bi_iter)
    {
        int count = bi_iter->second;

        if (bi_cc.count(count) > 0)
            ++(bi_cc[count]);
        else
            bi_cc[count] = 1;
    }

    double bN1 = bi_cc[1];
    double bN2 = bi_cc[2];
    double bN3 = bi_cc[3];
    double bN4 = bi_cc[4];

    double bY = bN1 / (bN1 + 2 * bN2);
    bD1 = 1 - 2 * bY*(bN2 / bN1);
    bD2 = 2 - 3 * bY*(bN3 / bN2);
    bD3plus = 3 - 4 * bY*(bN4 / bN3);

    if (!m_flag)
    {
        bD2 = bD1;
        bD3plus = bD1;
    }
}


double IQuadKneserNey::GetDiscountedQuadgramProb(int i, int j, int k, int l)
{
    iquadple v(i, j, k, l);
    itriple key(j, k, l);


    map<iquadple, int>& quadgram_counts = spec_mom->getQuadgramCounts();

    if (quadgram_counts.count(v) == 0)
        return 0;
    else
    {
        int qc = quadgram_counts[v];
        int tc = spec_mom->GetTHCount(key);

        double d_qc = qc;
        if (qc == 1)
        {
            d_qc -= qD1;
        }
        else if (qc == 2)
        {
            d_qc -= qD2;
        }
        else if (qc >= 3 && tc <= q_max_d_limit)
        {
            d_qc -= qD3plus;
        }

        double val = d_qc / tc;
        return val;
    }
}


double IQuadKneserNey::GetDiscountedTrigramProb(int i, int j, int k)
{
    vector<int> v = VectorPlus::CreateTriple(i, j, k);
    pair<int, int> p(j, k);
    double tc = (double)GetN1Prev(i, j, k);
    double bc = denom_t_n1prev->coeff(j, k);


    if (tc == 0.0)
        return 0.0;

    double d_tc = tc;
    if (tc == 1)
    {
        d_tc -= tD1;
    }
    else if (tc == 2)
    {
        d_tc -= tD2;
    }
    else if (tc >= 3 && tc <= t_max_d_limit)
    {
        d_tc -= tD3plus;
    }

    double val = d_tc / bc;
    return val;
}



double IQuadKneserNey::GetNonDiscTrigramProb(int i, int j, int k)
{
    vector<int> v = VectorPlus::CreateTriple(i, j, k);
    pair<int, int> p(j, k);

    double tc = (double)GetN1Prev(i, j, k);
    double bc = denom_t_n1prev->coeff(j, k);

    if (tc == 0.0)
        return 0.0;

    double val = tc / bc;
    return val;
}




double IQuadKneserNey::GetNonDiscBigramProb(int i, int j)
{
    double n1prev_c = (double)GetN1Prev(i, j);
    if (n1prev_c == 0.0)
        return 0;

    double denom = (*denom_b_n1prev)(j);
    double val = n1prev_c / denom;
    return val;
}


double IQuadKneserNey::GetDiscountedBigramProb(int i, int j)
{
    double n1prev_c = (double)GetN1Prev(i, j);
    double numer = 0.0;
    if (n1prev_c == 1)
    {
        numer = n1prev_c - bD1;
    }
    else if (n1prev_c == 2)
    {
        numer = n1prev_c - bD2;
    }
    else if (n1prev_c >= 3 && n1prev_c <= b_max_d_limit)
    {
        numer = n1prev_c - bD3plus;
    }

    double denom = (*denom_b_n1prev)(j);
    double val = numer / denom;
    return val;
}



double IQuadKneserNey::GetN1Prev(int i)
{
    if (u_n1prev_counts->count(i) > 0)
    {
        return (*u_n1prev_counts)[i];
    }
    else
        return 0;
}

double IQuadKneserNey::GetN1Prev(int i, int j)
{
    pair<int, int> p(i, j);

    if (b_n1prev_counts->count(p) > 0)
        return (*b_n1prev_counts)[p];
    else
        return 0;
}

double IQuadKneserNey::GetN1Prev(int i, int j, int k)
{
    itriple v(i, j, k);

    if (t_n1prev_counts->count(v) > 0)
        return (*t_n1prev_counts)[v];
    else
        return 0;
}

double IQuadKneserNey::GetProb(int i, int j, int k, int l)
{
    double qprob = GetDiscountedQuadgramProb(i, j, k, l);
    double tprob = GetDiscountedTrigramProb(i, j, k);
    double bprob = GetDiscountedBigramProb(i, j);
    double uprob = GetDiscountedUnigramProb(i);

    double tweight = GetKneserWeight(j, k, l);
    double bweight = GetKneserWeight(j, k);
    double uweight = GetKneserWeight(j);
    cout.flush();
    double total_bprob = bprob + uweight * uprob;
    double total_tprob = tprob + bweight * total_bprob;
    double total_prob = qprob + tweight * total_tprob;

    return total_prob;
}

double IQuadKneserNey::GetQTProb(int i, int j, int k, int l)
{
    double qprob = GetDiscountedQuadgramProb(i, j, k, l);
    double tprob = GetNonDiscTrigramProb(i, j, k);
    double tweight = GetKneserWeight(j, k, l);
    double total_prob = qprob + tweight * tprob;
    return total_prob;
}

//  double GetQTProb(int i, int j, int k, int l);
// double GetQTBProb(int i, int j, int k, int l);

double IQuadKneserNey::GetQTBProb(int i, int j, int k, int l)
{
    double qprob = GetDiscountedQuadgramProb(i, j, k, l);
    double tprob = GetDiscountedTrigramProb(i, j, k);
    double bprob = GetNonDiscBigramProb(i, j);

    double tweight = GetKneserWeight(j, k, l);
    double bweight = GetKneserWeight(j, k);

    double total_tprob = tprob + bweight * bprob;
    double total_prob = qprob + tweight * total_tprob;

    return total_prob;
}