/* Ankur Parikh
   IWittenBell.cpp

   see .hpp for class/function comments */



#include "IWittenBell.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include <vector>
#include <string>
#include <boost/algorithm/string.hpp>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <iostream>
#include <fstream>
#include "IOLibrary.hpp"

using namespace Eigen;
using namespace std;

//default constructor is used; 


using namespace std;

//constructor that initializes Moments
IWittenBell::IWittenBell(SpecMoments* spec_mom)
{
    this->spec_mom = spec_mom;
    b_n1plus_counts = new map<pair<int, int>, int>();
    u_n1plus_counts = new map<int, int>();

    //ComputeUCounts();
    //ComputeBCounts();
}




void IWittenBell::ComputeUCounts()
{
    assert(0);
    map<pair<int, int>, int>& b_counts = spec_mom->getBigramCounts();
    map<pair<int, int>, int>::iterator iter;

    map<int, set<int>*> u_n1plus_map;

    for (iter = b_counts.begin(); iter != b_counts.end(); ++iter)
    {
        pair<int, int> p = iter->first;

        // n1plus_set
        if (u_n1plus_map.count(p.second) == 0)
            u_n1plus_map[p.second] = new set<int>();

        u_n1plus_map[p.second]->insert(p.first);
    }


    map<int, set<int>*>::iterator plus_iter;

    for (plus_iter = u_n1plus_map.begin(); plus_iter != u_n1plus_map.end(); ++plus_iter)
    {
        int key = plus_iter->first;
        set<int>* val_set = plus_iter->second;
        int val = val_set->size();

        (*u_n1plus_counts)[key] = val;
        delete(val_set);
    }
}

void IWittenBell::ComputeBCounts()
{
    assert(0);
    map<itriple, int>& t_counts = spec_mom->getTrigramCounts();
    map<itriple, int>::iterator iter;

    map<pair<int, int>, set<int>*> b_n1plus_map;

    for (iter = t_counts.begin(); iter != t_counts.end(); ++iter)
    {
        itriple key = iter->first;
        pair<int, int> p(key.get<1>(), key.get<2>());

        // plus set
        if (b_n1plus_map.count(p) == 0)
            b_n1plus_map[p] = new set<int>();

        b_n1plus_map[p]->insert(key.get<0>());

    }

    // fill up plus and prev maps
    map<pair<int, int>, set<int>*>::iterator plus_iter;

    for (plus_iter = b_n1plus_map.begin(); plus_iter != b_n1plus_map.end(); ++plus_iter)
    {
        pair<int, int> key = plus_iter->first;
        set<int>* val_set = plus_iter->second;
        int val = val_set->size();

        (*b_n1plus_counts)[key] = val;
        delete(val_set);
    }
}


void IWittenBell::GetWeightVec(VectorXd& weights)
{
    assert(0);
    weights.resize(spec_mom->GetV());
    weights.fill(0);
    for (int c = 0; c < spec_mom->GetV(); ++c)
    {
        double n1plus = GetN1Plus(c);
        double uh_count = spec_mom->GetUHCount(c);

        if (n1plus == 0)
        {
            assert(uh_count == 0);
            continue;
        }

        double denom = uh_count + n1plus;

        weights(c) = (n1plus / denom);
    }

}

void IWittenBell::WittenBellBigram(MatrixXd& smoothed_P21)
{
    assert(0);
    VectorXd& unigrams = spec_mom->getUnigrams();
    smoothed_P21.resize(spec_mom->GetV(), spec_mom->GetV());

    for (int c = 0; c < spec_mom->GetV(); ++c)
    {
        double n1plus = GetN1Plus(c);
        double uh_count = spec_mom->GetUHCount(c);

        if (n1plus == 0)
        {
            assert(uh_count == 0);
            continue;
        }
        double denom = uh_count + n1plus;

        for (int r = 0; r < spec_mom->GetV(); ++r)
        {
            double e_cond_prob = spec_mom->GetCondProb(r, c);
            double pi = unigrams(c);

            double s_cond_prob = (uh_count / denom) * e_cond_prob + (n1plus / denom) * unigrams(r);
            double prob = s_cond_prob * pi;
            smoothed_P21(r, c) = prob;
        }
    }

}

double IWittenBell::GetBiWeight(int i)
{
    assert(0);
    double n1plus = GetN1Plus(i);
    double uh_count = spec_mom->GetUHCount(i);

    if (n1plus == 0)
    {
        assert(uh_count == 0);
        assert(0);
    }

    double denom = uh_count + n1plus;

    double w = (n1plus / denom);
    return w;
}

double IWittenBell::GetTriWeight(int i, int j)
{
    assert(0);
    //  if (j == 0)
    // {
    //	  return GetBiWeight(i);
    // }

    pair<int, int> p(i, j);
    double n1plus = GetN1Plus(p);
    double bh_count = spec_mom->GetBHCount(p);

    if (n1plus == 0)
    {
        if (bh_count > 0)
            assert(0);

        return 1;
    }

    double denom = bh_count + n1plus;
    double weight = (n1plus / denom);
    return weight;
}


double IWittenBell::GetN1Plus(int c)
{
    assert(0);
    if (u_n1plus_counts->count(c) > 0)
    {
        return (*u_n1plus_counts)[c];
    }
    else
        return 0;
}


double IWittenBell::GetN1Plus(pair<int, int> p)
{
    assert(0);
    if (b_n1plus_counts->count(p) > 0)
    {
        return (*b_n1plus_counts)[p];
    }
    else
        return 0;
}
