/* Ankur Parikh
   PLREWrapper.cpp

   see .hpp for class/function comments
   */

#include "PLREWrapper.hpp"
#include <assert.h>

using namespace std;
using namespace Eigen;


PLREWrapper::PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre, PLRE* tri_plre, QuadPLRE* quad_plre)
{
    this->spec_mom = spec_mom;
    this->bi_plre = bi_plre;
    this->tri_plre = tri_plre;
    this->quad_plre = quad_plre;

    quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
    triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
    pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
}

PLREWrapper::PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre, PLRE* tri_plre)
{
    this->spec_mom = spec_mom;
    this->bi_plre = bi_plre;
    this->tri_plre = tri_plre;
    this->quad_plre = NULL;

    quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
    triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
    pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
}

PLREWrapper::PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre)
{
    this->spec_mom = spec_mom;
    this->bi_plre = bi_plre;
    this->tri_plre = NULL;
    this->quad_plre = NULL;

    quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
    triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
    pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
}



double PLREWrapper::GetCondProb(const string& w, const string& wp, const string& wpp, const string& wppp)
{
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    int k = spec_mom->GetFinalID(wpp);
    int l = spec_mom->GetFinalID(wppp);

    if (quad_plre != NULL)
    {
        iquadple q(i, j, k, l);
        if (quad_cache->count(q) > 0)
            return (*quad_cache)[q];
        else
        {
            double val = GetCondProb(i, j, k, l);
            (*quad_cache)[q] = val;
            if (quad_cache->size() > 10000)
            {
                cerr << "cache cleared\n";
                quad_cache->clear();
            }
            return val;
        }
    }
    else
    {
        //	assert(0);
        return GetCondProb(w, wp, wpp);
    }
}

double PLREWrapper::GetCondProb(const string& w, const string& wp, const string& wpp)
{
    //	assert(wpp == start_marker);
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    int k = spec_mom->GetFinalID(wpp);
    itriple t(i, j, k);
    if (tri_plre != NULL)
    {
        if (triple_cache->count(t) > 0)
            return (*triple_cache)[t];
        else
        {
            double val = GetCondProb(i, j, k);
            (*triple_cache)[t] = val;
            if (triple_cache->size() > 10000)
            {
                cerr << "cache cleared\n";
                triple_cache->clear();
            }
            return val;
        }
    }
    else
    {
        return GetCondProb(w, wp);
    }
}

double PLREWrapper::GetCondProb(const string& w, const string& wp)
{
    //	assert(wp == start_marker);
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    pair<int, int> p(i, j);
    if (pair_cache->count(p) > 0)
        return (*pair_cache)[p];
    else
    {
        double val = GetCondProb(i, j);
        (*pair_cache)[p] = val;
        if (pair_cache->size() > 10000)
        {
            cerr << "cache cleared\n";
            pair_cache->clear();
        }

        return val;
    }
}


double PLREWrapper::GetCondProb(int i, int j, int k, int l)
{
    if (quad_plre != NULL)
    {
        double prob = quad_plre->GetCondProb(i, j, k, l);
        return prob;
    }
    else
    {
        return GetCondProb(i, j, k);
    }
}

double PLREWrapper::GetCondProb(int i, int j, int k)
{
    if (tri_plre != NULL)
    {
        double prob = tri_plre->GetCondProb(i, j, k);
        return prob;
    }
    else
    {
        return GetCondProb(i, j);
    }
}

double PLREWrapper::GetCondProb(int i, int j)
{
    double prob = bi_plre->GetBiCondProb(i, j);
    return prob;
}

