/* Ankur Parikh
   PowerFullBigram.cpp

   see .hpp for class/function comments */

#include "PowerFullBigram.hpp"
#include "EigenHelper.hpp"

using namespace Eigen;
using namespace std;


PowerFullBigram::PowerFullBigram()
{
    disc_bigram = NULL;
    denom = NULL;
    after_gamma = NULL;
}

PowerFullBigram::~PowerFullBigram()
{
    delete(disc_bigram);
    delete(denom);
    delete(after_gamma);
}

PowerFullBigram::PowerFullBigram(SparseMatrix<double>& bigram, SparseMatrix<double>& disc,
    VectorXd& after_gamma)
{
    disc_bigram = new SparseMatrix<double>();
    denom = new VectorXd();
    this->after_gamma = new VectorXd(after_gamma);

    *disc_bigram = bigram - disc;
    double min_val = EigenHelper::Min(*disc_bigram);
    if (min_val < -1e-10)
        assert(0);
    EigenHelper::ColSums(bigram, *denom);
}

double PowerFullBigram::LRDiscAt(int i, int j)
{
    double val = disc_bigram->coeff(i, j);
    return val;
}

double PowerFullBigram::CondLRDiscAt(int i, int j)
{
    double n = LRDiscAt(i, j);
    double d = denom->coeff(j);
    if (d == 0.0)
        return 0.0;

    double val = n / d;
    return val;
}

double PowerFullBigram::GammaAt(int j)
{
    if (denom->coeff(j) == 0.0)
    {
        //	assert(0);
        return 1.0;
    }
    return after_gamma->coeff(j) / denom->coeff(j);
}

