/* Ankur Parikh
   PowerFullBigram.hpp

   powered yet full (and thus sparse) bigram in the power low rank ensemble */

#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerBigram.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;

class PowerFullBigram : public PowerBigram
{
private:

    SparseMatrix<double>* disc_bigram; // discounted bigram counts
    VectorXd* denom; // denominators (column sums) to help compute conditionals
    VectorXd* after_gamma; // leftover weight after discounting

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & disc_bigram;
        ar & denom;
        ar & after_gamma;
    }

public:

    PowerFullBigram();
    ~PowerFullBigram();
    PowerFullBigram(SparseMatrix<double>& bigram, SparseMatrix<double>& disc, VectorXd& after_gamma);

    // returns discounted joint probability P_d(i,j)
    double LRDiscAt(int i, int j);

    // returns discounted conditional probability P_d(i|j)
    double CondLRDiscAt(int i, int j);

    // returns leftover weight for history j
    double GammaAt(int j);
};