/* Ankur Parikh
   PowerFullQuadgram.cpp

   see .hpp for class/function comments */

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerFullQuadgram.hpp"
#include "EigenHelper.hpp"

using namespace Eigen;
using namespace std;



PowerFullQuadgram::PowerFullQuadgram()
{
    counts = NULL;
    disc_quad_mat = NULL;
    denom_ten = NULL;
    after_gamma_ten = NULL;
}


PowerFullQuadgram::~PowerFullQuadgram()
{
    delete(disc_quad_mat);
    delete(denom_ten);
    delete(after_gamma_ten);
}

PowerFullQuadgram::PowerFullQuadgram(SMat4Tensor<int>* counts, SparseMatrix<double>& disc,
    VectorXd& after_gamma)
{
    this->counts = counts;
    SparseMatrix<int>& count_mat = counts->GetMat();
    // compute colsums;
    denom_ten = new VectorXd();
    EigenHelper::ColSums(count_mat, *denom_ten);

    disc_quad_mat = new SparseMatrix<double>();

    *disc_quad_mat = count_mat.cast<double>() - disc;
    double min_val = EigenHelper::Min(*disc_quad_mat);
    if (min_val < -1e-10)
        assert(0);

    counts->ResetSpMat();
    after_gamma_ten = new VectorXd(after_gamma);
}

double PowerFullQuadgram::LRDiscAt(int i, int j, int k, int l)
{
    int alt_index = counts->ComputeAltIndex(j, k, l);
    if (alt_index == -1)
        return 0;
    double val = disc_quad_mat->coeff(i, alt_index);
    return val;
}

double PowerFullQuadgram::CondLRDiscAt(int i, int j, int k, int l)
{
    int alt_index = counts->ComputeAltIndex(j, k, l);
    if (alt_index == -1)
        return 0;
    double n = LRDiscAt(i, j, k, l);
    double d = denom_ten->coeff(alt_index);
    if (d == 0.0)
        return 0;

    double val = n / d;
    return val;
}

double PowerFullQuadgram::GammaAt(int j, int k, int l)
{
    int alt_index = counts->ComputeAltIndex(j, k, l);
    if (alt_index == -1)
        return 1;
    double g = after_gamma_ten->coeff(alt_index);
    double d = denom_ten->coeff(alt_index);
    if (d == 0.0)
        return 1.0;
    else
        return g / d;
}