/* Ankur Parikh
   PowerFullQuadgram.hpp

   powered yet full (and thus sparse) quadgram in the power low rank ensemble */

#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "SMat4Tensor.hpp"
#include "PowerQuadgram.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;

class PowerFullQuadgram : public PowerQuadgram
{
private:

    SMat4Tensor<int>* counts; // fourth order tensor counts
    SparseMatrix<double>* disc_quad_mat; // discounted counts

    VectorXd* denom_ten; // denominators to compute conditional probabilities
    VectorXd* after_gamma_ten; // leftover weight after discounting

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & counts;
        ar & disc_quad_mat;
        ar & denom_ten;
        ar & after_gamma_ten;
    }

public:

    PowerFullQuadgram();
    ~PowerFullQuadgram();
    PowerFullQuadgram(SMat4Tensor<int>* counts, SparseMatrix<double>& disc, VectorXd& after_gamma);

    // returns discounted joint probability P_d(i,j,k,l)
    double LRDiscAt(int i, int j, int k, int l);

    // returns discounted conditional probability P_d(i|j,k,l)
    double CondLRDiscAt(int i, int j, int k, int l);

    // leftoverweight gamma for history (j,k,l)
    double GammaAt(int j, int k, int l);
};