/* Ankur Parikh
   PowerFullTrigram.cpp

   see .hpp for class/function comments
   */


#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerFullTrigram.hpp"
#include "EigenHelper.hpp"

using namespace Eigen;
using namespace std;



PowerFullTrigram::PowerFullTrigram()
{
    counts = NULL;
    disc_tri_mat = NULL;
    denom_mat = NULL;
    after_gamma_mat = NULL;
}


PowerFullTrigram::~PowerFullTrigram()
{
    delete(disc_tri_mat);
    delete(denom_mat);
    delete(after_gamma_mat);
}

PowerFullTrigram::PowerFullTrigram(SMat3Tensor<int>* counts, SparseMatrix<double>& disc,
    VectorXd& after_gamma)
{
    this->counts = counts;
    SparseMatrix<int>& count_mat = counts->GetMat();
    // compute colsums;
    VectorXd denom_vec;
    EigenHelper::ColSums(count_mat, denom_vec);

    disc_tri_mat = new SparseMatrix<double>();
    denom_mat = new SparseMatrix<double>(disc.rows(), disc.rows());
    this->after_gamma_mat = new SparseMatrix<double>(disc.rows(), disc.rows());

    *disc_tri_mat = count_mat.cast<double>() - disc;
    double min_val = EigenHelper::Min(*disc_tri_mat);
    if (min_val < -1e-10)
        assert(0);

    // rearrange into sparse matrix
    vector<triplet> d_triplets;
    vector<triplet> g_triplets;

    for (int i = 0; i < after_gamma.size(); ++i)
    {
        pair<int, int> p = counts->ComputeReverseAltIndex(i);

        double d = denom_vec.coeff(i);
        double g = after_gamma(i);

        d_triplets.push_back(triplet(p.first, p.second, d));
        g_triplets.push_back(triplet(p.first, p.second, g));
    }

    denom_mat->setFromTriplets(d_triplets.begin(), d_triplets.end());
    after_gamma_mat->setFromTriplets(g_triplets.begin(), g_triplets.end());

    counts->ResetSpMat();
}

double PowerFullTrigram::LRDiscAt(int i, int j, int k)
{
    int alt_index = counts->ComputeAltIndex(j, k);
    if (alt_index == -1)
        return 0;
    double val = disc_tri_mat->coeff(i, alt_index);
    return val;
}

double PowerFullTrigram::CondLRDiscAt(int i, int j, int k)
{
    double n = LRDiscAt(i, j, k);
    double d = denom_mat->coeff(j, k);
    if (d == 0.0)
        return 0;

    double val = n / d;
    return val;
}

double PowerFullTrigram::GammaAt(int j, int k)
{
    double g = after_gamma_mat->coeff(j, k);
    double d = denom_mat->coeff(j, k);
    if (d == 0.0)
        return 1.0;
    else
        return g / d;
}