/* Ankur Parikh
   PowerFullTrigram.hpp

   powered yet full (and thus sparse) trigram in the power low rank ensemble */

#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerTrigram.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;

class PowerFullTrigram : public PowerTrigram
{
private:

    SMat3Tensor<int>* counts; // counts
    SparseMatrix<double>* disc_tri_mat; // discounted counts in matrix form

    SparseMatrix<double>* denom_mat; // denominators to compute conditional probabilities
    SparseMatrix<double>* after_gamma_mat; // leftover weight after discounting

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & counts;
        ar & disc_tri_mat;
        ar & denom_mat;
        ar & after_gamma_mat;
    }

public:

    PowerFullTrigram();
    ~PowerFullTrigram();
    PowerFullTrigram(SMat3Tensor<int>* counts, SparseMatrix<double>& disc, VectorXd& after_gamma);

    // returns discounted joint probability P_d(i,j,k)
    double LRDiscAt(int i, int j, int k);

    // returns discounted conditional probability P_d(i |j,k)
    double CondLRDiscAt(int i, int j, int k);

    // leftoverweight gamma for history (j,k)
    double GammaAt(int j, int k);
};