/* Ankur Parikh
   PowerLRBigram.hpp

   low rank powered bigram used in power low rank ensemble

   stored as B_lr = L * R, where B_lr(i,j) = L(i) * R(j)

   where L is called l_cpt, and R is called R_cpt
   */

#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerBigram.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;


class PowerLRBigram : public PowerBigram
{
private:

    VectorXd* denom; // denominator for fast computation of conditional
    MatrixXd* l_cpt; // L in B = L * R
    MatrixXd* r_cpt; // R in B = L * R
    MatrixXd* real_l_cpt; // can be ignored

    VectorXd* after_gamma; // leftover weight

    double f; // rank parameter

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & denom;
        ar & l_cpt;
        ar & r_cpt;
        ar & after_gamma;
        ar & f;
    }

public:

    // constructor / initializers
    PowerLRBigram();
    PowerLRBigram(SparseMatrix<double>& power_counts, SparseMatrix<double>& disc, VectorXd& before_gamma,
        VectorXd& after_gamma, double f);
    void Initialize(SparseMatrix<double>& power_counts, SparseMatrix<double>& disc, VectorXd& before_gamma,
        VectorXd& after_gamma, double f);

    ~PowerLRBigram();

    // returns discounted joint probability P_d(i,j)
    double LRDiscAt(int i, int j);

    // returns discounted conditional probability P_d(i|j)
    double CondLRDiscAt(int i, int j);

    // returns leftover weight
    double GammaAt(int j);

    // these functions can be ignored, they just return top words for visualization
    vector<pair<int, double> > TopLeftIndices(int index, int num_top);

    vector<pair<int, double> > TopRightIndices(int index, int num_top);
};
