/* Ankur Parikh
   PowerLRQuadgram.hpp

   low rank powered quadgram used in power low rank ensemble

   stored as dictionary of left / right condensed matrices where

   Q_lr(i,j,k, l) = < l_cpts[(j,k)]->at(i), r_cpts([j,k])->at(l) > (where <,> is inner product)

   */


#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "SMat4Tensor.hpp"
#include "SQuadMatrixArray.hpp"
#include "PowerQuadgram.hpp"
#include "CondensedDMat.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;


class PowerLRQuadgram : public PowerQuadgram
{
private:

    map<pair<int, int>, CondensedDMat*>* r_cpts; // right side low rank matrix indexed by slice (j,k)
    map<pair<int, int>, CondensedDMat*>* l_cpts; // left side low rank matrix indexed by slice (j,k)

    SQuadMatrixArray<double>* count_array; // counts reformatted

    VectorXd* denom_ten; // denominators (j,k,l) for fast computation of conditional
    VectorXd* after_gamma_ten; // leftover weights

    SMat4Tensor<int>* counts; //counts 

    double f; // rank

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & counts;
        ar & r_cpts;
        ar & l_cpts;
        ar & count_array;
        ar & denom_ten;
        ar & after_gamma_ten;
        ar & f;
    }

public:

    // constructor / initializers
    PowerLRQuadgram();
    PowerLRQuadgram(SMat4Tensor<int>* counts, SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
        VectorXd& before_gamma, VectorXd& after_gamma, double f);
    void Initialize(SMat4Tensor<int>* counts, SparseMatrix<double>& power_counts,
        SparseMatrix<double>& disc, VectorXd& before_gamma, VectorXd& after_gamma, double f);

    ~PowerLRQuadgram();

    // returns discounted joint probability P_d(i,j,k,l)
    double LRDiscAt(int i, int j, int k, int l);

    // returns discounted conditional probability P_d(i|j,k,l)
    double CondLRDiscAt(int i, int j, int k, int l);

    // returns leftover weight j,k,l
    double GammaAt(int j, int k, int l);
};