/* Ankur Parikh
   PowerLRTrigram.hpp

   low rank powered trigram used in power low rank ensemble

   stored as vector of left / right condensed matrices wherex

   T_lr(i,j,k) = < l_cpts->at(j)->at(i), r_cpts->at(j)->at(k) > (where <,> is inner product)

   */

#pragma once

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include "IKneserNey.hpp"
#include "SMat3Tensor.hpp"
#include "PowerTrigram.hpp"
#include "CondensedDMat.hpp"
#include "eigen_boost_serialization.hpp"
#include <boost/unordered_map.hpp>

using namespace Eigen;
using namespace std;


class PowerLRTrigram : public PowerTrigram
{
private:

    vector<CondensedDMat*>* r_cpts; // stores right side matrices of low rank structure
    vector<CondensedDMat*>* l_cpts; // stores left side matrices of low rank structure


    SMatrixArray<double>* count_array; // counts reformatted into matrix array

    VectorXd* denoms; // denominators for fast conditional computation
    VectorXd* after_gammas; // leftover weights

    SMat3Tensor<int>* counts; // counts
    //	map<itriple,double>* cache;
    int K; // rank
    double f; // rank via factor 

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & r_cpts;
        ar & l_cpts;
        ar & denoms;
        ar & after_gammas;
        ar & counts;
        ar & K;
        ar & f;
    }

public:

    // constructor / initializers
    PowerLRTrigram();
    PowerLRTrigram(SMat3Tensor<int>* counts, SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
        VectorXd& before_gamma, VectorXd& after_gamma, int K, double f);
    void Initialize(SMat3Tensor<int>* counts, SparseMatrix<double>& power_counts,
        SparseMatrix<double>& disc, VectorXd& before_gamma, VectorXd& after_gamma, int K, double f);

    ~PowerLRTrigram();

    // returns discounted joint probability P_d(i,j,k)
    double LRDiscAt(int i, int j, int k);

    // returns discounted conditional probability P_d(i | j,k)
    double CondLRDiscAt(int i, int j, int k);

    // returns leftover weight for history (j,k)
    double GammaAt(int j, int k);
};