/* Ankur Parikh
   QuadLMWrapper.cpp

   see *hpp for class/function comments
   */


#include "QuadLMWrapper.hpp"


void QuadLMWrapper::RunQuadKneserNeyPLRE(IKNWrapper& kn, PLREWrapper& plre,
    SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag)
{
    int numExamples = data.size();

    VectorXd& unigrams = mom.getUnigrams();
    VectorXd likelihoods_kn(numExamples);
    VectorXd likelihoods_e(numExamples);
    likelihoods_kn.fill(0);
    likelihoods_e.fill(0);
    int wc = 0;

    for (int i = 0; i < numExamples; ++i) // loop through sentences
    {
        vector<int>& sentence = *(data.at(i));

        double ll_kn = 0; // kneser ney log likelihood
        double ll_e = 0; // plre log likelihood

        int prev_w = -1;
        int prev_prev_w = -1;
        int prev_prev_prev_w = -1;

        for (int ch = 0; ch < sentence.size(); ++ch) // loop through words in sentence
        {
            if (sentence[ch] < 0) // end of sentence
                assert(0);

            int w = sentence[ch];
            if (sentence[ch] == std::numeric_limits<int>::max()) //oov word
            {
                w = mom.GetOOVIndex();

            }

            // start symbol
            if (ch == 0)
            {
                prev_w = w;
                continue;
            }

#pragma omp atomic
            wc++; //increment word count

            if (ch == 1)
            {
                double kn_prob = kn.GetCondProb(w, prev_w);
                double plre_prob = plre.GetCondProb(w, prev_w);

                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }
                ll_kn += log(kn_prob) / log(10.0);
                ll_e += log(plre_prob) / log(10.0);

            }
            else if (ch == 2)
            {
                double kn_prob = kn.GetCondProb(w, prev_w, prev_prev_w);
                double plre_prob = plre.GetCondProb(w, prev_w, prev_prev_w);

                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }

                ll_kn += log(kn_prob) / log(10.0);
                ll_e += log(plre_prob) / log(10.0);
            }
            else
            {
                double kn_prob = kn.GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);
                double plre_prob = plre.GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);
                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }

                ll_kn += log(kn_prob) / log(10.0);
                ll_e += log(plre_prob) / log(10.0);
            }

            prev_prev_prev_w = prev_prev_w;
            prev_prev_w = prev_w;
            prev_w = w;
        }

#pragma omp critical
        {
            likelihoods_kn(i) = ll_kn;
            likelihoods_e(i) = ll_e;
        }
    }

    double val_kn = likelihoods_kn.sum();
    double val_e = likelihoods_e.sum();

    // print out log likelihoods
    cout << "word count: " << wc << "\n"; // word count
    cout << "log-likelihood(base10)_ModIntKneserNey: " << val_kn << "\n";
    cout << "log-likelihood(base10)_PLRE:  " << val_e << "\n\n";
    cout << "Perplexity_ModIntKneserNey: " << pow(10, -1 * val_kn / wc) << "\n";
    cout << "Perplexity_PLRE:  " << pow(10, -1 * val_e / wc) << "\n\n";
    cout.flush();
}

void QuadLMWrapper::RunQuadKneserNey(IKNWrapper& kn, SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag)
{
    int numExamples = data.size();

    VectorXd& unigrams = mom.getUnigrams();
    VectorXd likelihoods_kn(numExamples);
    VectorXd likelihoods_e(numExamples);
    likelihoods_kn.fill(0);
    likelihoods_e.fill(0);
    int wc = 0;
#pragma omp parallel for
    for (int i = 0; i < numExamples; ++i) // loop through sentences
    {
        vector<int>& sentence = *(data.at(i));

        double ll_kn = 0; // kneser ney log likelihood

        int prev_w = -1;
        int prev_prev_w = -1;
        int prev_prev_prev_w = -1;

        for (int ch = 0; ch < sentence.size(); ++ch) // loop through words in sentence
        {
            if (sentence[ch] < 0) // end of sentence
                assert(0);

            int w = sentence[ch];
            if (sentence[ch] == std::numeric_limits<int>::max()) //oov word
            {
                w = mom.GetOOVIndex();
            }


            // start symbol
            if (ch == 0)
            {
                prev_w = w;
                continue;
            }

#pragma omp atomic
            wc++; // increment word count

            if (ch == 1)
            {
                double kn_prob = kn.GetCondProb(w, prev_w);
                ll_kn += log(kn_prob) / log(10.0);
            }
            else if (ch == 2)
            {
                double kn_prob = kn.GetCondProb(w, prev_w, prev_prev_w);
                ll_kn += log(kn_prob) / log(10.0);
            }
            else
            {
                double kn_prob = kn.GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);


                ll_kn += log(kn_prob) / log(10.0);

            }

            prev_prev_prev_w = prev_prev_w;
            prev_prev_w = prev_w;
            prev_w = w;
        }

#pragma omp critical
        {
            likelihoods_kn(i) = ll_kn;
        }
    }

    double val_kn = likelihoods_kn.sum();

    cout << "val_kn: " << val_kn << "\n";
    cout << "wc: total" << wc << "\n"; // word count
    cout.flush();
}



void QuadLMWrapper::RunQuadPLRE(PLREWrapper& plre, SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag)
{
    int numExamples = data.size();

    VectorXd& unigrams = mom.getUnigrams();
    VectorXd likelihoods_kn(numExamples);
    VectorXd likelihoods_e(numExamples);
    likelihoods_kn.fill(0);
    likelihoods_e.fill(0);
    int wc = 0;

    for (int i = 0; i < numExamples; ++i) // loop through sentences
    {
        vector<int>& sentence = *(data.at(i));
        if (i % 10000 == 0)
        {
            cout << "test: " << i << "\n";
            cout.flush();
        }
        double ll_e = 0; // plre log likelihood

        int prev_w = -1;
        int prev_prev_w = -1;
        int prev_prev_prev_w = -1;

        for (int ch = 0; ch < sentence.size(); ++ch) // loop through words in sentence
        {
            if (sentence[ch] < 0) // end of sentence
                assert(0);

            int w = sentence[ch];
            if (sentence[ch] == std::numeric_limits<int>::max()) //oov word
            {
                w = mom.GetOOVIndex();
            }


            // start symbol
            if (ch == 0)
            {
                prev_w = w;
                continue;
            }

#pragma omp atomic
            wc++; // increment word count

            if (ch == 1)
            {
                double plre_prob = plre.GetCondProb(w, prev_w);

                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }
                ll_e += log(plre_prob) / log(10.0);
            }
            else if (ch == 2)
            {
                double plre_prob = plre.GetCondProb(w, prev_w, prev_prev_w);

                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }

                ll_e += log(plre_prob) / log(10.0);
            }
            else
            {
                double plre_prob = plre.GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);

                if (plre_prob == 0.0 && incomplete_flag)
                {
                    continue;
                }

                ll_e += log(plre_prob) / log(10.0);
            }

            prev_prev_prev_w = prev_prev_w;
            prev_prev_w = prev_w;
            prev_w = w;
        }

#pragma omp critical
        {
            likelihoods_e(i) = ll_e;
        }
    }

    double val_e = likelihoods_e.sum();

    cout << "val_plre: " << val_e << "\n";
    cout << "wc: total" << wc << "\n"; // word count
    cout.flush();
}
