/* Ankur Parikh
   QuadLMWrapper.hpp

   class that evaluates the different language models and computes perplexity

   vector< vector<int>* >& data = dataset, each element A of the vector is a vector of word ids for line A in the dataset
   incomplete_flag = false if whole model has been learned, true if only higher order n-grams are learned
   */



# pragma once

#include <vector>
#include "Tensor.hpp"
#include "SMatrixArray.hpp"
#include "EigenHelper.hpp"
#include <complex>
#include <Eigen/Dense>
#include <Eigen/SVD>
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"
#include "PowerLRTrigram.hpp"
#include "PLRE.hpp"
#include "QuadPLRE.hpp"
#include "IQuadKneserNey.hpp"
#include "IKNWrapper.hpp"
#include "PLREWrapper.hpp"

using namespace Eigen;

using namespace std;

class QuadLMWrapper
{
public:

    // run both methods and compute perplexity
    static void RunQuadKneserNeyPLRE(IKNWrapper& kn, PLREWrapper& plre,
        SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag);

    // run only Kneser Ney and compute perplexity
    static void RunQuadKneserNey(IKNWrapper& kn, SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag);

    // run only PLRE and compute perplexity
    static void RunQuadPLRE(PLREWrapper& plre, SpecMoments& mom, vector< vector<int>* >& data, bool incomplete_flag);
};