/* Ankur Parikh
   SMat4Tensor.hpp

   matricized fourth order tensor, where one index is on the rows and 3 are on the columns
   */


#define EIGEN_YES_I_KNOW_SPARSE_MODULE_IS_NOT_STABLE_YET

# pragma once

#include <vector>
#include "Tensor.hpp"
#include <Eigen/Sparse>
#include "Tensor.hpp"
#include <map>
#include <iostream>
#include <fstream>
#include "eigen_boost_serialization.hpp"
#include "serialize_tuple.hpp"

using namespace std;
using namespace Eigen;

template <class T>
class SMat4Tensor
{
private:

    SparseMatrix<T>* sp_ten_alt; // matricized tensor

    map<itriple, int>* triple_to_index; // tensor index triple to column index 
    vector< Triplet<T> >* alt_triplet_list; // matricized tensor in form of sparse vector of triplets
    vector<int>* dims; // dimensions of tensor

    bool isInitialized; // flag
    bool isFilled;

    int curr_alt_index;

    int mode; // which mode is on the rows

    void ComputeOffsets();

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & sp_ten_alt;
        ar & triple_to_index;
        //	ar & index_to_triple;
        //	ar & alt_triplet_list;
        ar & dims; // dimensions of tensor

        ar & isInitialized; // flag
        ar & isFilled;

        ar & curr_alt_index;

        ar & mode;
    }

public:

    SMat4Tensor();
    SMat4Tensor(vector<int>& dim_array, int mode);
    ~SMat4Tensor();

    void ResetSpMat()
    {
        delete(sp_ten_alt);
        sp_ten_alt = new SparseMatrix<T>(1, 1);
    }


    // intialization
    void Initialize(vector<int>& dim_array, int mode);

    int GetMode() { return mode; }

    map<itriple, int>& GetTriple2IndexMap() { return *triple_to_index; }

    // basic accessors
    vector<int>& Dims() { return *dims; }
    int Dim(int index) { return (*dims)[index]; }

    Eigen::SparseVector<double> GetCol(int j, int k, int l)
    {
        itriple t(j, k, l);
        int index = ComputeAltIndex(j, k, l);
        if (index == -1)
        {
            Eigen::SparseVector<double> v(1);
            return v;
        }
        else
        {
            return sp_ten_alt->col(index);
        }
    }

    // access elements of tensor
    T At(int i, int j, int k, int l);

    void StartFill(int reserve_count);
    void EndFill();

    // set elements of tensor
    void Set(int i, int j, int k, int l, T val);

    SparseMatrix<T>& GetMat() { return *sp_ten_alt; }

    int ComputeAltIndex(int i, int j, int k);
};

template <class T>
// default constructor
SMat4Tensor<T>::SMat4Tensor()
{
    sp_ten_alt = NULL;

    triple_to_index = NULL;
    //	index_to_triple = NULL;
    alt_triplet_list = NULL;
    dims = NULL; // dimensions of tensor


    isInitialized = false; // flag
    isFilled = false;

    curr_alt_index = -1;

    mode = -1;
}

template <class T>
SMat4Tensor<T>::SMat4Tensor(vector<int>& dim_array, int mode)
{
    alt_triplet_list = NULL;
    isFilled = false;
    isInitialized = false;
    Initialize(dim_array, mode);
}

// destructor
template <class T>
SMat4Tensor<T>::~SMat4Tensor()
{
    delete(sp_ten_alt);
    delete(triple_to_index);
    //	delete(index_to_triple);
    //	delete(alt_triplet_list);
    delete(dims);
}

// initialize
template <class T>
void SMat4Tensor<T>::Initialize(vector<int>& dims, int mode)
{
    assert(!isInitialized);

    assert(dims.size() == 4);

    sp_ten_alt = new SparseMatrix<T>();
    triple_to_index = new map<itriple, int>();
    //	index_to_triple = new map<int, itriple >();
    this->dims = new vector<int>(dims);

    curr_alt_index = 0;

    isInitialized = true;
    this->mode = mode;
    assert(mode == 0);
}


// access elements of tensor
template <class T>
T SMat4Tensor<T>::At(int i, int j, int k, int l)
{
    assert(isInitialized);
    assert(isFilled);

    if (mode == 0)
    {
        assert(isFilled);
        int alt_index = ComputeAltIndex(j, k, l);
        if (alt_index < 0)
            return 0;
        else
            return sp_ten_alt->coeff(i, alt_index);
    }
    /*	else if (mode == 3)
        {
        int alt_index = ComputeAltIndex(i,j, k);
        if (alt_index < 0)
        return 0;
        else
        return sp_ten_alt->coeff(alt_index, k);
        }*/
    else
    {
        assert(0);
    }
}

template <class T>
void SMat4Tensor<T>::StartFill(int reserve_count)
{
    assert(!isFilled);
    alt_triplet_list = new vector< Triplet<T> >();
    alt_triplet_list->reserve(reserve_count);
}

template <class T>
void SMat4Tensor<T>::EndFill()
{
    if (mode == 0)
    {
        sp_ten_alt->resize(dims->at(0), curr_alt_index);
    }
    /*	else if (mode == 2)
        {
        sp_ten_alt->resize(curr_alt_index, dims->at(2));
        }*/
    else
    {
        assert(0);
    }

    sp_ten_alt->setFromTriplets(alt_triplet_list->begin(), alt_triplet_list->end());
    //cout << sp_ten_alt->rows() << "\t" << sp_ten_alt->cols() << "\n";
    isFilled = true;
    delete(alt_triplet_list);
}


template <class T>
void SMat4Tensor<T>::Set(int i, int j, int k, int l, T val)
{
    assert(!isFilled);
    assert(alt_triplet_list != NULL);

    if (mode == 0)
    {
        int alt_col_index = ComputeAltIndex(j, k, l);
        if (alt_col_index == -1)
        {
            itriple v(j, k, l);
            (*triple_to_index)[v] = curr_alt_index;
            //	(*index_to_triple)[curr_alt_index] = v;
            curr_alt_index++;
        }
        alt_col_index = ComputeAltIndex(j, k, l);
        assert(alt_col_index >= 0);

        alt_triplet_list->push_back(Triplet<T>(i, alt_col_index, val));
    }
    /*	else if (mode == 2)
        {
        int alt_col_index = ComputeAltIndex(i,j);
        if (alt_col_index == -1)
        {
        pair<int, int> p(i,j);
        (*pair_to_index)[p] = curr_alt_index;
        (*index_to_pair)[curr_alt_index] = p;
        curr_alt_index++;
        }
        alt_col_index = ComputeAltIndex(i,j);
        assert(alt_col_index >= 0);

        alt_triplet_list->push_back(triplet(alt_col_index, k, val));
        }*/
    else
    {
        assert(0);
    }
}
template <class T>
int SMat4Tensor<T>::ComputeAltIndex(int j, int k, int l)
{
    itriple v(j, k, l);
    if (triple_to_index->count(v) > 0)
        return (*triple_to_index)[v];
    else
        return -1;
}

