/* Ankur Parikh
   SpecMoments.cpp

   see *hpp file for class/function descriptions
   */

#include "SpecMoments.hpp"
#include "SMatrixArray.hpp"
#include "VectorPlus.hpp"
#include <vector>
#include <string>
#include <boost/algorithm/string.hpp>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <iostream>
#include <fstream>
#include "IOLibrary.hpp"
#include <math.h>

using namespace Eigen;
using namespace std;

//default constructor is used; 

using namespace std;

//constructor that initializes Moments
SpecMoments::SpecMoments(string fn, string oov_arg,
    string start_marker, string end_marker, int cores){

    bigram_cmat = NULL;
    trigram_cmat = NULL;
    quadgram_cmat = NULL;

    th_counts = NULL;
    bh_counts = NULL;
    uh_counts = NULL;

    unigrams = new VectorXd();
    // bigram_cmat = new SparseMatrix<int>(); 
    //	trigram_cmat = new SMat3Tensor<int>();
    //	quadgram_cmat = new SMat4Tensor<int>();
    this->num_cores = cores;
    rev_ngram_map = new map<int, string>();
    oov_marker = oov_arg;

    ngram_map = new map<string, int>();
    unigram_counts = new map<int, int>();
    bigram_counts = new map<pair<int, int>, int>();
    trigram_counts = new map<itriple, int>();
    quadgram_counts = new map<iquadple, int>();

    //	th_counts = new map<itriple,int>();
    //	bh_counts = new map<pair<int, int>, int>();
    //	uh_counts = new map<int, int>();

    curr_w_index = 0;

    readUnigrams(fn, start_marker, end_marker);
    cout << "finished reading unigrams \n";

    for (int i = 0; i < 3; ++i)
    {
        if (i == 0)
        {
            readBigrams(fn);
            cout << "finished reading bigrams \n";
        }
        if (i == 1)
        {
            readMatTrigrams(fn);
            cout << "finished reading trigrams \n";
        }
        if (i == 2)
        {
            readMatQuadgrams(fn);
            cout << "finished reading quadgrams \n";
        }
    }
}

void SpecMoments::readUnigrams(string fn, string start_marker, string end_marker){

    ifstream in_file;
    in_file.open(fn.c_str());
    string line;

    int normalizer = 0;
    num_sents = 0;

    boost::unordered_map<int, int> temp_ucounts;

    bool oov_found = false;
    bool smarker_found = false;
    bool emarker_found = false;

    while (getline(in_file, line))
    {
        boost::trim(line);
        vector<string> strs;
        boost::split(strs, line, boost::is_any_of(" "));


        // if (num_sents % 10000 == 0)
        // 		cout << "uni: " << num_sents << "\n";
        num_sents++;
        for (int i = 0; i < strs.size(); ++i)
        {
            string c = strs[i];


            if (GetID(c) == std::numeric_limits<int>::max())
            {
                (*ngram_map)[c] = curr_w_index++;
                (*rev_ngram_map)[curr_w_index - 1] = c;

                if (c == oov_marker)
                    oov_found = true;

                if (c == end_marker)
                    emarker_found = true;

                /*  if (c[0] == '<' && c[c.size()-1] == '>' && c.find("gunit") != string::npos)
                  {
                  assert(oov_marker == "");
                  oov_marker = c;
                  } */
            }

            if (c == start_marker)
            {
                smarker_found = true;
                continue;
            }

            int c_index = (*ngram_map)[c];
            if (temp_ucounts.count(c_index) == 0)
                temp_ucounts[c_index] = 1;
            else
                ++(temp_ucounts[c_index]);

            normalizer++;
        }
    }

    if (normalizer == 0)
        ASSERT_WITH_MESSAGE(0, "training file either not found or empty");

    if (!oov_found)
        ASSERT_WITH_MESSAGE(0, "oov marker not found in the training data");

    if (!smarker_found)
        ASSERT_WITH_MESSAGE(0, "start marker not found in the training data");

    if (!emarker_found)
        ASSERT_WITH_MESSAGE(0, "end marker not found in the training data");

    //	assert(oov_marker.size() > 0 && ngram_map->count(oov_marker) > 0);
    // put unordered_map back in the std::map
    boost::unordered_map<int, int>::iterator iter;
    for (iter = temp_ucounts.begin(); iter != temp_ucounts.end(); ++iter)
    {
        (*unigram_counts)[iter->first] = iter->second;
    }


    unigrams->resize(ngram_map->size());
    for (int i = 0; i < unigrams->size(); ++i)
    {
        (*unigrams)(i) = ((*unigram_counts)[i] + 0.0) / (normalizer + 0.0);
    }
    in_file.close();
}



void SpecMoments::readBigrams(string fn)
{
    int temp_index = 0;

    vector< map<pair<int, int>, int>* > b_count_vec;

    for (int i = 0; i < num_cores; ++i)
        b_count_vec.push_back(new map<pair<int, int>, int>());

    //	#pragma omp parallel for
    for (int n = 0; n < num_cores; ++n)
    {
        string fn_n = fn + "-" + ToString(n);
        if (n == 0 && num_cores == 1)
            fn_n = fn;

        ifstream in_file;
        in_file.open(fn_n.c_str());
        string line;

        map<pair<int, int>, int>* b_counts = b_count_vec[n];

        while (getline(in_file, line))
        {
            boost::trim(line);
            vector<string> strs;
            boost::split(strs, line, boost::is_any_of(" "));
            temp_index++;
            //	if (temp_index % 10000 == 0)
            //		cout << "bi: " << temp_index << "\n";

            for (int i = 0; i < ((int)strs.size() - 1); ++i)
            {
                int index1 = (*ngram_map)[strs[i + 1]];
                int index2 = (*ngram_map)[strs[i]];

                pair<int, int> p(index1, index2);

                if (b_counts->count(p) == 0)
                    (*b_counts)[p] = 1;
                else
                    ++((*b_counts)[p]);
            }
        }
        in_file.close();
    }

    // put unordered_map back in the std::map

    // now we need to merge all the maps into the quadgram counts
    int normalizer = 0;
    for (int i = 0; i < num_cores; ++i)
    {
        map<pair<int, int>, int>& b_counts = *(b_count_vec[i]);
        map<pair<int, int>, int>::iterator bi_iter;
        for (bi_iter = b_counts.begin(); bi_iter != b_counts.end(); ++bi_iter)
        {
            pair<int, int> p = bi_iter->first;
            int c = bi_iter->second;
            normalizer += c;
            if (bigram_counts->count(p) == 0)
                (*bigram_counts)[p] = c;
            else
                (*bigram_counts)[p] = (*bigram_counts)[p] + c;
        }
        delete(b_count_vec[i]);
    }
}

void SpecMoments::readMatTrigrams(string fn){

    int temp_index = 0;

    vector< map<itriple, int>* > t_count_vec;

    for (int i = 0; i < num_cores; ++i)
        t_count_vec.push_back(new map<itriple, int>());

    //	#pragma omp parallel for
    for (int n = 0; n < num_cores; ++n)
    {
        string fn_n = fn + "-" + ToString(n);
        if (n == 0 && num_cores == 1)
            fn_n = fn;
        ifstream in_file;
        in_file.open(fn_n.c_str());
        string line;

        map<itriple, int>* t_counts = t_count_vec[n];

        while (getline(in_file, line))
        {
            boost::trim(line);
            vector<string> strs;
            boost::split(strs, line, boost::is_any_of(" "));
            temp_index++;
            //	if (temp_index % 10000 == 0)
            //			cout << "tri: " << temp_index << "\n";

            for (int i = 0; i < ((int)strs.size() - 2); ++i)
            {
                int index1 = (*ngram_map)[strs[i + 2]];
                int index2 = (*ngram_map)[strs[i + 1]];
                int index3 = (*ngram_map)[strs[i]];

                itriple key(index1, index2, index3);
                //	vector<int> key = VectorPlus::CreateTriple(index1, index2, index3);

                // trigram counts
                if (t_counts->count(key) == 0)
                    (*t_counts)[key] = 1;
                else
                    ++((*t_counts)[key]);
            }
        }
        in_file.close();
    }

    // now we need to merge all the maps into the quadgram counts
    //	int normalizer = 0;
    for (int i = 0; i < num_cores; ++i)
    {
        map<itriple, int>& t_counts = *(t_count_vec[i]);
        map<itriple, int>::iterator ti_iter;
        for (ti_iter = t_counts.begin(); ti_iter != t_counts.end(); ++ti_iter)
        {
            itriple key = ti_iter->first;
            int c = ti_iter->second;
            pair<int, int> p(key.get<1>(), key.get<2>());
            if (trigram_counts->count(key) == 0)
                (*trigram_counts)[key] = c;
            else
                (*trigram_counts)[key] = (*trigram_counts)[key] + c;
        }
        delete(t_count_vec[i]);
    }
}


void SpecMoments::readMatQuadgrams(string fn){

    int temp_index = 0;

    vector< map<iquadple, int>* > q_count_vec;

    for (int i = 0; i < num_cores; ++i)
        q_count_vec.push_back(new map<iquadple, int>());

    //	#pragma omp parallel for
    for (int n = 0; n < num_cores; ++n)
    {
        string fn_n = fn + "-" + ToString(n);
        if (n == 0 && num_cores == 1)
            fn_n = fn;
        ifstream in_file;
        in_file.open(fn_n.c_str());
        string line;

        map<iquadple, int>* q_counts = q_count_vec[n];

        while (getline(in_file, line))
        {
            boost::trim(line);
            vector<string> strs;
            boost::split(strs, line, boost::is_any_of(" "));
            temp_index++;
            //	if (temp_index % 10000 == 0)
            //			cout << "quad: " << temp_index << "\n";

            for (int i = 0; i < ((int)strs.size() - 3); ++i)
            {
                int index1 = (*ngram_map)[strs[i + 3]];
                int index2 = (*ngram_map)[strs[i + 2]];
                int index3 = (*ngram_map)[strs[i + 1]];
                int index4 = (*ngram_map)[strs[i]];

                iquadple key(index1, index2, index3, index4);

                if (q_counts->count(key) == 0)
                    (*q_counts)[key] = 1;
                else
                    ++((*q_counts)[key]);
            }
        }
        in_file.close();
    }

    // now we need to merge all the maps into the quadgram counts
    //	int normalizer = 0;
    int max_qcount = 0;
    int max_thcount = 0;
    for (int i = 0; i < num_cores; ++i)
    {
        map<iquadple, int>& q_counts = *(q_count_vec[i]);
        map<iquadple, int>::iterator qi_iter;
        for (qi_iter = q_counts.begin(); qi_iter != q_counts.end(); ++qi_iter)
        {
            iquadple key = qi_iter->first;
            int c = qi_iter->second;
            itriple t(key.get<1>(), key.get<2>(), key.get<3>());
            if (quadgram_counts->count(key) == 0)
                (*quadgram_counts)[key] = c;
            else
                (*quadgram_counts)[key] = (*quadgram_counts)[key] + c;

            if ((*quadgram_counts)[key] > max_qcount)
            {
                max_qcount = (*quadgram_counts)[key];
            }

            if ((*quadgram_counts)[key] > max_qcount)
            {
                max_qcount = (*quadgram_counts)[key];
            }
        }
        delete(q_count_vec[i]);
    }
}


void SpecMoments::FillHCounts()
{
    th_counts = new map<itriple, int>();
    bh_counts = new map<pair<int, int>, int>();
    uh_counts = new map<int, int>();

    map<pair<int, int>, int>::iterator bi_iter;
    for (bi_iter = bigram_counts->begin(); bi_iter != bigram_counts->end(); ++bi_iter)
    {
        pair<int, int> p = bi_iter->first;
        int c = bi_iter->second;

        if (uh_counts->count(p.second) == 0)
            (*uh_counts)[p.second] = c;
        else
            (*uh_counts)[p.second] = (*uh_counts)[p.second] + c;
    }

    map<itriple, int>::iterator ti_iter;
    for (ti_iter = trigram_counts->begin(); ti_iter != trigram_counts->end(); ++ti_iter)
    {
        itriple key = ti_iter->first;
        int c = ti_iter->second;
        pair<int, int> p(key.get<1>(), key.get<2>());

        if (bh_counts->count(p) == 0)
            (*bh_counts)[p] = c;
        else
            (*bh_counts)[p] = (*bh_counts)[p] + c;
    }

    map<iquadple, int>::iterator qi_iter;
    for (qi_iter = quadgram_counts->begin(); qi_iter != quadgram_counts->end(); ++qi_iter)
    {
        iquadple key = qi_iter->first;
        int c = qi_iter->second;
        itriple t(key.get<1>(), key.get<2>(), key.get<3>());


        if (th_counts->count(t) == 0)
            (*th_counts)[t] = c;
        else
            (*th_counts)[t] = (*th_counts)[t] + c;
    }

}


void SpecMoments::FillTensors()
{
    bigram_cmat = new SparseMatrix<int>();
    trigram_cmat = new SMat3Tensor<int>();
    quadgram_cmat = new SMat4Tensor<int>();

    bigram_cmat->resize(ngram_map->size(), ngram_map->size());
    vector<triplet> triplet_vec;
    triplet_vec.reserve(bigram_counts->size());
    map<pair<int, int>, int>::iterator biter;

    cout << "\nstarting bi tensor fill\n";
    for (biter = bigram_counts->begin(); biter != bigram_counts->end(); ++biter)
    {
        pair<int, int> p = biter->first;
        int c = ((double)(biter->second));
        triplet_vec.push_back(triplet(p.first, p.second, c));
    }

    bigram_cmat->setFromTriplets(triplet_vec.begin(), triplet_vec.end());

    vector<int> tdims = VectorPlus::CreateTriple(curr_w_index, curr_w_index, curr_w_index);
    trigram_cmat->Initialize(tdims, 0);

    trigram_cmat->StartFill(trigram_counts->size());

    map<itriple, int>::iterator titer;
    int temp_index = 0;
    cout << "\nstarting tri tensor fill\n";
    for (titer = trigram_counts->begin(); titer != trigram_counts->end(); ++titer)
    {
        //		if (temp_index % 10000 == 0)
        // 		cout << "tri2: " << temp_index << "\n";

        itriple v = titer->first;
        pair<int, int> p(v.get<1>(), v.get<2>());
        double val = (*unigrams)(v.get<0>());

        trigram_cmat->Set(v.get<0>(), v.get<1>(), v.get<2>(), (double)(titer->second));
    }

    trigram_cmat->EndFill();

    cout << "\nstarting quad tensor fill\n";
    vector<int> dims = VectorPlus::CreateQuartet(curr_w_index, curr_w_index, curr_w_index, curr_w_index);

    quadgram_cmat->Initialize(dims, 0);

    quadgram_cmat->StartFill(quadgram_counts->size());

    map<iquadple, int>::iterator iter;
    temp_index = 0;
    for (iter = quadgram_counts->begin(); iter != quadgram_counts->end(); ++iter)
    {
        iquadple v = iter->first;
        //	double prob = ((double)(iter->second)) / normalizer;

        quadgram_cmat->Set(v.get<0>(), v.get<1>(), v.get<2>(), v.get<3>(), (double)(iter->second));
    }

    quadgram_cmat->EndFill();
}

double SpecMoments::GetCondProb(int i, int j)
{

    pair<int, int> p(i, j);
    if (bigram_counts->count(p) > 0)
    {
        int numer = (*bigram_counts)[p];
        int denom = (*uh_counts)[j];
        if (j == 0)
            denom = num_sents;

        return (numer + 0.0) / (denom + 0.0);
    }
    else
    {
        return 0;
    }
}

double SpecMoments::GetCondProb(int i, int j, int k)
{
    //  if (k == 0)
    //	  return GetCondProb(i,j);

    itriple t(i, j, k);
    pair<int, int> p(j, k);
    if (trigram_counts->count(t) > 0)
    {
        int numer = (*trigram_counts)[t];
        int denom = (*bh_counts)[p];

        return (numer + 0.0) / (denom + 0.0);
    }
    else
    {
        return 0;
    }
}
