/* Ankur Parikh
   VectorPlus.hpp

   Helper class for additional function for std::vector
   */

# pragma once

#include <vector>
#include <set>

using namespace std;

class VectorPlus
{
public:

    static void Copy(vector<int>& result_vec, vector<int>& vec);// copy vec into result_vec
    static void Reverse(vector<int>& result_vec, vector<int>& vec); // reverse vec into result_vec
    static void Concat(vector<int>& result_vec, vector<int>& vec_A, vector<int>& vec_B); // result_vec = vec_A concatenated with vec_B
    static void Concat(vector<int>& vec_A, vector<int>& vec_B); // vec_A = vec_A concatenate vec_B
    static void Unique(vector<int>& result_vec, vector<int>& vec); // get unique elements of vec into result_vec

    static bool Equals(vector<int>& vec_A, vector<int>& vec_B); // tests if two vectors are equal
    static bool Contains(vector<int>& vec, int element); // test if vec contains element
    static bool IsSubset(vector<int>& vec_A, vector<int>& vec_B); // test if vec_B is a subset of vec_A	
    static int Find(vector<int>& vec, int element); // find first position of element in vec
    static void Find(vector<int>& result_vec, vector<int>& vec, int element); // find all indices that element appears in vec

    static int Product(vector<int>& vec); // product of all elements of vec
    static int Sum(vector<int>& vec); // sum of all elements of vec
    static double Sum(vector<double>& vec); // sum of all elements of vec
    static int AbsSum(vector<int>& vec); // absolute sum of vec
    static void Add(vector<int>& sum_vec, vector<int>& vec_A, vector<int>& vec_B); // sum_vec = vec_A + vec_B

    static void Union(vector<int>& result_vec, vector<int>& vec_A, vector<int>& vec_B); // union of vec_A and vec_B into result_vec
    static void Union(vector<int>& vec_A, vector<int>& vec_B); // in place union of vec_A and vec_B, result in vec_A
    static void SetDiff(vector<int>& result_vec, vector<int>& vec_A, vector<int>& vec_B); // result_vec = vec_A \setminus vec_B
    static void SetDiff(vector<int>& result_vec, vector<int>& vec_A, set<int>& set_B); // result_vec = vec_A \setminus set_B
    static void Intersect(vector<int>& result_vec, vector<int>& vec_A, vector<int>& vec_B); // result_vec = vec_A \setminus set_B
    static void Intersect(vector<int>& result_vec, vector<int>& vec_A, set<int>& set_B); // result_vec = vec_A \intersect set_B
    static void Subset(vector<int>& result_vec, vector<int>& vec, vector<int>& indices); // find subset of vec indicated by indices, put in result_vec
    static void CSubset(vector<int>& result_vec, vector<int>& vec, vector<int>& indices); // find complement of subset of vec indicated by indices, put in result_vec
    static void Subset(vector<int>& result_vec, vector<int>& vec, int start_index, int end_index); // find complement of subset of vec from start_index to end_index, put in result_vec

    static double Min(vector<double>& vec); // minimum element of vec
    static double Max(vector<double>& vec); // maximum element of vec
    static int MinIndex(vector<int>& vec); // minimum index of vec
    static int MaxIndex(vector<int>& vec); // maximum index of vec

    static vector<int> CreateSingleton(int val); // vector with one element: val
    static vector<int> CreatePair(int val1, int val2); // vector with two elements: val1, val2
    static vector<int> CreateTriple(int val1, int val2, int val3);
    static vector<int> CreateQuartet(int val1, int val2, int val3, int val4);
    static vector<double> CreateSingleton(double val); // vector with one element: val
    static vector<double> CreatePair(double val1, double val2); // vector with two elements: val1, val2
    static vector<double> CreateTriple(double val1, double val2, double val3); // vector with three elements: val1, val2, val3
    static vector<double> CreateQuartet(double val1, double val2, double val3, double val4); // vector with four elements: val1, val2, val3, val4
    static void Seq(vector<int>& result_vec, int start, int step, int end); // result_vec is a sequence of ints from start to end spaced by step
};
