// serialization helper, downloaded from StackOverflow


#pragma once 

#include <vector>
#include <Eigen/Sparse>
#include <Eigen/Dense>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/split_free.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/split_member.hpp>
#include <boost/functional/hash.hpp>
#include "serialize_tuple.hpp"

using namespace std;
using namespace Eigen;


namespace boost {
    namespace serialization {
        class access;
    }
}

namespace boost {
    namespace serialization {

        template <class Archive, typename _Scalar, int _Rows, int _Cols, int _Options, int _MaxRows, int _MaxCols>
        void save(Archive & ar, const Eigen::Matrix<_Scalar, _Rows, _Cols, _Options, _MaxRows, _MaxCols> & m, const unsigned int version)
        {
            int rows = m.rows(), cols = m.cols();
            ar & rows;
            ar & cols;
            ar & boost::serialization::make_array(m.data(), rows*cols);
        }

        template <class Archive, typename _Scalar, int _Rows, int _Cols, int _Options, int _MaxRows, int _MaxCols>
        void load(Archive & ar, Eigen::Matrix<_Scalar, _Rows, _Cols, _Options, _MaxRows, _MaxCols> & m, const unsigned int version)
        {
            int rows, cols;
            ar & rows;
            ar & cols;
            m.resize(rows, cols);
            ar & boost::serialization::make_array(m.data(), rows*cols);
        }

        template <class Archive, typename _Scalar, int _Rows, int _Cols, int _Options, int _MaxRows, int _MaxCols>
        void serialize(Archive & ar, Eigen::Matrix<_Scalar, _Rows, _Cols, _Options, _MaxRows, _MaxCols> & m, const unsigned int version)
        {
            split_free(ar, m, version);
        }

        template <class Archive>
        void save(Archive & ar, const SparseMatrix<int>& m, const unsigned int version)
        {
            int rows = m.rows();
            int cols = m.cols();
            int nnz = m.nonZeros();

            ar & rows;
            ar & cols;
            ar & nnz;

            int innerSize = m.innerSize();
            int outerSize = m.outerSize();

            for (int i = 0; i < outerSize; ++i)
            {

                for (typename Eigen::SparseMatrix<int>::InnerIterator it(m, i); it; ++it)
                {
                    int r = it.row();
                    int c = it.col();
                    int val = it.value();
                    ar & r;
                    ar & c;
                    ar & val;
                }
            }
        }
        template <class Archive>
        void load(Archive & ar, SparseMatrix<int>& m, const unsigned int version)
        {
            int rows, cols, nnz;
            ar & rows;
            ar & cols;
            ar & nnz;
            m.resize(rows, cols);
            vector< Triplet<int> > triplets;
            for (int i = 0; i < nnz; ++i)
            {
                int r;
                int c;
                double val;
                ar & r;
                ar & c;
                ar & val;

                triplets.push_back(Triplet<int>(r, c, val));
            }
            m.setFromTriplets(triplets.begin(), triplets.end());
        }

        template <class Archive>
        void serialize(Archive & ar, SparseMatrix<int>& m, const unsigned int version)
        {
            split_free(ar, m, version);
        }

        template <class Archive>
        void save(Archive & ar, const SparseMatrix<int, RowMajor>& m, const unsigned int version)
        {
            int rows = m.rows();
            int cols = m.cols();
            int nnz = m.nonZeros();

            ar & rows;
            ar & cols;
            ar & nnz;

            int innerSize = m.innerSize();
            int outerSize = m.outerSize();

            for (int i = 0; i < outerSize; ++i)
            {

                for (typename Eigen::SparseMatrix<int, RowMajor>::InnerIterator it(m, i); it; ++it)
                {
                    int r = it.row();
                    int c = it.col();
                    double val = it.value();
                    ar & r;
                    ar & c;
                    ar & val;
                }
            }
        }
        template <class Archive>
        void load(Archive & ar, SparseMatrix<int, RowMajor>& m, const unsigned int version)
        {
            int rows, cols, nnz;
            ar & rows;
            ar & cols;
            ar & nnz;
            m.resize(rows, cols);
            vector< Triplet<double> > triplets;
            for (int i = 0; i < nnz; ++i)
            {
                int r;
                int c;
                double val;
                ar & r;
                ar & c;
                ar & val;

                triplets.push_back(Triplet<double>(r, c, val));
            }
            m.setFromTriplets(triplets.begin(), triplets.end());
        }

        template <class Archive>
        void serialize(Archive & ar, SparseMatrix<int, RowMajor>& m, const unsigned int version)
        {
            split_free(ar, m, version);
        }

        template <class Archive>
        void save(Archive & ar, const SparseMatrix<double>& m, const unsigned int version)
        {
            int rows = m.rows();
            int cols = m.cols();
            int nnz = m.nonZeros();

            ar & rows;
            ar & cols;
            ar & nnz;

            int innerSize = m.innerSize();
            int outerSize = m.outerSize();

            for (int i = 0; i < outerSize; ++i)
            {

                for (typename Eigen::SparseMatrix<double>::InnerIterator it(m, i); it; ++it)
                {
                    int r = it.row();
                    int c = it.col();
                    double val = it.value();
                    ar & r;
                    ar & c;
                    ar & val;
                }
            }
        }
        template <class Archive>
        void load(Archive & ar, SparseMatrix<double>& m, const unsigned int version)
        {
            int rows, cols, nnz;
            ar & rows;
            ar & cols;
            ar & nnz;
            m.resize(rows, cols);
            vector< Triplet<double> > triplets;
            for (int i = 0; i < nnz; ++i)
            {
                int r;
                int c;
                double val;
                ar & r;
                ar & c;
                ar & val;

                triplets.push_back(Triplet<double>(r, c, val));
            }
            m.setFromTriplets(triplets.begin(), triplets.end());
        }

        template <class Archive>
        void serialize(Archive & ar, SparseMatrix<double>& m, const unsigned int version) {
            split_free(ar, m, version);
        }

        template <class Archive>
        void save(Archive & ar, const SparseMatrix<double, RowMajor>& m, const unsigned int version)
        {
            int rows = m.rows();
            int cols = m.cols();
            int nnz = m.nonZeros();

            ar & rows;
            ar & cols;
            ar & nnz;

            int innerSize = m.innerSize();
            int outerSize = m.outerSize();

            for (int i = 0; i < outerSize; ++i)
            {

                for (typename Eigen::SparseMatrix<double, RowMajor>::InnerIterator it(m, i); it; ++it)
                {
                    int r = it.row();
                    int c = it.col();
                    double val = it.value();
                    ar & r;
                    ar & c;
                    ar & val;
                }
            }
        }
        template <class Archive>
        void load(Archive & ar, SparseMatrix<double, RowMajor>& m, const unsigned int version)
        {
            int rows, cols, nnz;
            ar & rows;
            ar & cols;
            ar & nnz;
            m.resize(rows, cols);
            vector< Triplet<double> > triplets;
            for (int i = 0; i < nnz; ++i)
            {
                int r;
                int c;
                double val;
                ar & r;
                ar & c;
                ar & val;

                triplets.push_back(Triplet<double>(r, c, val));
            }
            m.setFromTriplets(triplets.begin(), triplets.end());
        }

        template <class Archive>
        void serialize(Archive & ar, SparseMatrix<double, RowMajor>& m, const unsigned int version)
        {
            split_free(ar, m, version);
        }

        template <class Archive, typename _Scalar>
        void save(Archive & ar, const Eigen::Triplet<_Scalar> & m, const unsigned int version)
        {
            ar & m.row();
            ar & m.col();
            ar & m.value();
        }

        template <class Archive, typename _Scalar>
        void load(Archive & ar, Eigen::Triplet<_Scalar> & m, const unsigned int version)
        {
            int row, col;
            _Scalar value;
            ar & row;
            ar & col;
            ar & value;
            m = Eigen::Triplet<_Scalar>(row, col, value);
        }

        template <class Archive, typename _Scalar>
        void serialize(Archive & ar, Eigen::Triplet<_Scalar> & m, const unsigned int version)
        {
            split_free(ar, m, version);
        }

    }
}

