/* Ankur Parikh
  ex_plre.cpp

  Shows how to test an already trained PLRE model with some examples. */

#include "Tensor.hpp"
# include <string>
#include "IOLibrary.hpp"
#include "VectorPlus.hpp"
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"
#include "EM.hpp"
#include "QuadPLRE.hpp"
#include "PLRE.hpp"
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include "eigen_boost_serialization.hpp"
#include <boost/serialization/export.hpp>
#include "QuadLMWrapper.hpp"


using namespace Eigen;

int main(int argc, char *argv[])
{
    string model_fn = "";

    for (int i = 1; i < argc; i = i + 2)
    {
        string arg_type = argv[i];
        string arg_val = argv[i + 1];

        if (arg_type == "--model")
        {
            model_fn = arg_val;
        }
        else
        {
            ASSERT_WITH_MESSAGE(0, "command line argument not valid");
        }
    }

    cout << "model name: " << model_fn << "\n";
    ASSERT_WITH_MESSAGE(model_fn.size() > 0, "no model name given");

    // test archive
    std::ifstream ifs(model_fn.c_str(), std::ios::binary);
    boost::archive::binary_iarchive ari(ifs);
    SpecMoments* moments_ar;
    IKNWrapper* kn_ar;
    PLREWrapper* plre_ar;

    cout << "\n\n----------Loading model from archive----------\n";
    ari  & moments_ar;
    cout << "moments loaded\n";
    ari & kn_ar;
    cout << "KN loaded\n";
    ari & plre_ar;
    cout << "PLRE loaded\n";
    cout << "Done loading model from archive\n";

    // test
    cout << "\n\n----------Starting Example Phrases----------" << "\n";

    string w = "green";
    string prev_w = "is";
    string prev_prev_w = "house";
    string prev_prev_prev_w = "the";

    cout << "Example phrase:\n";
    cout << prev_prev_prev_w << " " << prev_prev_w << " " << prev_w << " " << w << "\n\n";

    // this gets the PLRE probabilities (NOT in log scale) for order 4, order 3, order 2 models
    // respectively

    // P(w | prev_w, prev_prev_w, prev_prev_prev_w)
    double prob1_plre = plre_ar->GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);

    // P(w | prev_w, prev_prev_w)
    double prob2_plre = plre_ar->GetCondProb(w, prev_w, prev_prev_w);

    // P(w | prev_w)
    double prob3_plre = plre_ar->GetCondProb(w, prev_w);

    cout << "PLRE example probabilities: " << prob1_plre << " " << prob2_plre << " " << prob3_plre << "\n";

    // this gets the KN probabilities (NOT in log scale) for order 4, order 3, order 2 models
    // respectively

    // P(w | prev_w, prev_prev_w, prev_prev_prev_w)
    double prob1_kn = kn_ar->GetCondProb(w, prev_w, prev_prev_w, prev_prev_prev_w);

    // P(w | prev_w, prev_prev_w)
    double prob2_kn = kn_ar->GetCondProb(w, prev_w, prev_prev_w);

    // P(w | prev_w)
    double prob3_kn = kn_ar->GetCondProb(w, prev_w);

    cout << "KN example probabilities: " << prob1_kn << " " << prob2_kn << " " << prob3_kn << "\n";

    cout.flush();

    delete(moments_ar);
    delete(kn_ar);
    delete(plre_ar);

    return 0;
}