/* Ankur Parikh
  test_plre.cpp

  Shows how to test an already trained PLRE model. See README for details */

#include "Tensor.hpp"
# include <string>
#include "IOLibrary.hpp"
#include "VectorPlus.hpp"
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"
#include "EM.hpp"
#include "QuadPLRE.hpp"
#include "PLRE.hpp"
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include "eigen_boost_serialization.hpp"
#include <boost/serialization/export.hpp>
#include "QuadLMWrapper.hpp"


using namespace Eigen;

int main(int argc, char *argv[])
{
    string model_fn = "";
    string test_fn = "";

    for (int i = 1; i < argc; i = i + 2)
    {
        string arg_type = argv[i];
        string arg_val = argv[i + 1];

        if (arg_type == "--model")
        {
            model_fn = arg_val;
        }

        else if (arg_type == "--test")
        {
            test_fn = arg_val;
        }

        else
        {
            ASSERT_WITH_MESSAGE(0, "command line argument not valid");
        }
    }

    cout << "model name: " << model_fn << "\n";
    cout << "test file: " << test_fn << "\n";
    ASSERT_WITH_MESSAGE(model_fn.size() > 0, "no model name given");
    ASSERT_WITH_MESSAGE(test_fn.size() > 0, "no test file given");


    // test archive
    std::ifstream ifs(model_fn.c_str(), std::ios::binary);
    boost::archive::binary_iarchive ari(ifs);
    SpecMoments* moments_ar;
    IKNWrapper* kn_ar;
    PLREWrapper* plre_ar;

    cout << "\n\n----------Loading model from archive----------\n";
    ari  & moments_ar;
    cout << "moments loaded\n";
    ari & kn_ar;
    cout << "KN loaded\n";
    ari & plre_ar;
    cout << "PLRE loaded\n";
    cout << "Done loading model from archive\n";

    // test
    cout << "\n\n----------Starting Test----------" << "\n";
    cout.flush();




    vector< vector<int>* > test_data;
    GetBigTestMatrix(test_data, test_fn, *moments_ar);

    QuadLMWrapper::RunQuadKneserNeyPLRE(*kn_ar, *plre_ar,
        *moments_ar, test_data, false);

    delete(moments_ar);
    delete(kn_ar);
    delete(plre_ar);

    return 0;
}